import asyncio
import logging
import subprocess  # nosec
from argparse import Namespace
from pathlib import Path

import typer
from alembic import command, script
from alembic.config import Config
from alembic.util.exc import AutogenerateDiffsDetected
from sqlalchemy import text

from brickworks.core import db
from brickworks.core.constants import BASE_DIR, CORE_DIR
from brickworks.core.db import get_db_url
from brickworks.core.module_loader import load_models

logger = logging.getLogger(__name__)


db_app = typer.Typer()


@db_app.command()
def drop(schema: str) -> None:
    """Drop the database schema."""
    asyncio.run(_drop_schema(schema))


@db_app.command()
def upgrade(message: str | None = None) -> None:
    """Create and apply database migrations."""
    typer.echo("Upgrading database")
    if has_changes():
        make_migration(message)

    _migrate()


@db_app.command()
def migrate(schema: str = "") -> None:
    """Run database migrations."""
    typer.echo("Running database migrations")
    _migrate(schema)


@db_app.command()
def downgrade(levels: int = 1) -> None:
    """Downgrade database migrations."""
    typer.echo("Downgrading database")
    command.downgrade(get_config(), f"-{levels}")

    # delete the migration files
    migration_dir = Path(BASE_DIR) / "migrations"
    migration_files = sorted([f for f in migration_dir.iterdir() if f.suffix == ".py"], reverse=True)
    for i in range(levels):
        if i < len(migration_files):
            migration_files[i].unlink()
        else:
            break


@db_app.command()
def squash(force: bool = False) -> None:
    """
    Squashes the database migrations of the current branch.
    Will not work on main or master branch unless --force is used.
    This is useful for cleaning up the migration history and reducing the number of migrations.
    """
    typer.echo("Merging database migrations")
    branch = get_current_git_branch()
    if not force and branch in ("main", "master"):
        # This is a safety check to prevent squashing migrations on the main branch,
        # which is usually a bad idea because squashing on main would break deployed instances
        typer.echo("Cannot squash migrations on main or master branch. Use --force to override.")
        return

    levels = count_branch_revisions(branch)
    typer.echo(f"Squashing {levels} migrations")
    if levels == 0:
        typer.echo("No migrations to squash")
        return
    downgrade(levels)
    upgrade()


def count_branch_revisions(branch: str) -> int:
    """Count the number of revisions in the given branch."""
    sc = script.ScriptDirectory.from_config(get_config())
    revisions = sc.walk_revisions()
    count = 0
    for rev in revisions:
        if rev.path.endswith(branch + ".py"):
            count += 1
        else:
            # stop counting when we reach a revision that doesn't belong to the branch
            return count
    return count


def get_config() -> Config:
    config = Config(str(Path(CORE_DIR) / "alembic.ini"))
    config.set_main_option("sqlalchemy.url", get_db_url())
    config.set_main_option("script_location", str(Path(CORE_DIR) / "migration"))
    config.set_main_option("version_locations", str(Path(BASE_DIR) / "migrations"))
    return config


def _migrate(schema: str = "") -> None:
    """Run database migrations."""
    config = get_config()
    x = ["mode=migrate"]
    if schema:
        x.append(f"schema={schema}")
    config.cmd_opts = Namespace(x=x)
    command.upgrade(config, "head")


def make_migration(message: str | None = None) -> None:
    """Create a new database migration."""
    load_models()

    if not message:
        message = get_current_git_branch()

    config = get_config()
    config.cmd_opts = Namespace(autogenerate=True, message=message)

    command.revision(config, autogenerate=True, message=message)


def has_changes() -> bool:
    """
    Check if there are any migrations to be made.
    Returns true if there are unmigrated changes.
    """
    load_models()
    config = get_config()
    try:
        command.check(config=config)
    except AutogenerateDiffsDetected:
        return True
    return False


def get_current_git_branch() -> str:
    try:
        # Run the git command to get the current branch name
        result = subprocess.run(  # nosec
            ["git", "rev-parse", "--abbrev-ref", "HEAD"], capture_output=True, text=True, check=True
        )
        # Return the branch name (strip out any extra whitespace)
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        typer.echo("Failed to get current git branch. Make sure you are in a git repository.")
        return ""


async def _drop_schema(schema: str) -> None:
    """Drop the database schema."""
    typer.echo(f"Dropping database schema: {schema}")
    async with db(commit_on_exit=True):
        await db.session.execute(text(f'DROP SCHEMA IF EXISTS "{schema}" CASCADE'))



"""
OpenID Connect authentication using Authlib and FastAPI security.
To get a token use fastapi_oauth2 as a dependency in your route.
e.g.:
    @app.get("/login")
    async def login(token: str = fastapi_oauth2):
        user = await authlib_oauth.auth0.parse_id_token(token)
        return {"user": user}
"""

from authlib.integrations.starlette_client import OAuth, StarletteOAuth2App
from pydantic import AnyUrl, BaseModel, EmailStr

from brickworks.core.models.user_model import UserModel
from brickworks.core.settings import settings

authlib_oauth = OAuth()


def _get_oidc_client() -> StarletteOAuth2App:
    client = authlib_oauth.register(
        name=settings.OIDC_PROVIDER_NAME,
        client_id=settings.OIDC_CLIENT_ID,
        client_secret=settings.OIDC_CLIENT_SECRET,
        server_metadata_url=settings.OIDC_DISCOVERY_URL,
        client_kwargs={"scope": "openid profile email"},
    )
    if not client:
        raise ValueError("OIDC client was not registered.")
    return client


oidc_client: StarletteOAuth2App = _get_oidc_client()


class OIDCAddress(BaseModel):
    """
    OIDC Address Claim.
    https://openid.net/specs/openid-connect-core-1_0.html#StandardClaims
    """

    formatted: str | None = None
    street_address: str | None = None
    locality: str | None = None
    region: str | None = None
    postal_code: str | None = None
    country: str | None = None


class OIDCUserInfo(BaseModel):
    """
    OIDC Standard Claims.
    https://openid.net/specs/openid-connect-core-1_0.html#StandardClaims

    Note: The standard claim does not require the email claim to be present,
    however, we require it in our implementation - otherwise the user cannot be created.
    """

    sub: str

    email: EmailStr

    email_verified: bool | None = None
    name: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    middle_name: str | None = None
    nickname: str | None = None
    preferred_username: str | None = None
    profile: AnyUrl | None = None
    picture: AnyUrl | None = None
    website: AnyUrl | None = None

    gender: str | None = None
    birthdate: str | None = None
    zoneinfo: str | None = None
    locale: str | None = None
    phone_number: str | None = None
    phone_number_verified: bool | None = None
    address: OIDCAddress | None = None
    updated_at: int | None = None


async def create_user_from_userinfo(userinfo: OIDCUserInfo) -> UserModel:
    """
    Create a user from the userinfo dictionary returned by the OIDC provider.
    This is a placeholder implementation. In a real application, you would want to create a user in your database.
    """

    user = await UserModel(
        sub=userinfo.sub,
        email=str(userinfo.email),
        given_name=userinfo.given_name or "",
        family_name=userinfo.family_name or "",
        name=userinfo.name or f"{userinfo.given_name} {userinfo.family_name}",
        email_verified=userinfo.email_verified or False,
        phone_number=userinfo.phone_number or "",
        phone_number_verified=userinfo.phone_number_verified or False,
        locale=userinfo.locale or "en",
    ).persist()
    return user


async def get_or_create_user_from_userinfo(userinfo: OIDCUserInfo) -> UserModel:
    """
    Get or create a user from the userinfo dictionary returned by the OIDC provider.
    This is a placeholder implementation. In a real application, you would want to create a user in your database.
    """
    user = await UserModel.get_one_or_none(sub=userinfo.sub, _apply_policies=False)
    if not user:
        user = await create_user_from_userinfo(userinfo)
    return user



from datetime import UTC, datetime


def now_utc() -> datetime:
    return datetime.now(UTC)


def convert_datetime_to_iso_8601_with_z_suffix(dt: datetime) -> str:
    # convert dt to utc if it is timezone aware
    if dt.tzinfo is not None:
        dt = dt.astimezone(tz=None)
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")



import logging
from collections.abc import AsyncIterable
from typing import Any

import pytest
from asgi_lifespan import LifespanManager
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient

from brickworks.core import db
from brickworks.core.auth.authcontext import AuthContext
from brickworks.core.server import create_app

_app = create_app(for_testing=True)

logger = logging.getLogger(__name__)


TestApp = AsyncIterable[FastAPI]


@pytest.fixture
async def app() -> TestApp:
    """
    Create a FastAPI instance.

    1. Populate database with test data only once
    2. Create a FastAPI instance
    3. Execute lifespan cycle
    4. Create a global transaction to wrap each test.

       * DBTransactionMiddleware only wraps each REST API endpoint,
         but not the test itself.
       * The global force_rollback provided by Encode/Databases doesn't play well
         with inner transactions and cause issues.
    """
    async with LifespanManager(_app), db(), AuthContext():
        yield _app
        await db.session.rollback()


@pytest.fixture
async def client(app: FastAPI) -> AsyncIterable[AsyncClient]:
    async def clear_session(_: Any) -> None:  # noqa: ANN401
        # we need to make sure that subsequent requests don't have any data from the previous request
        await db.session.flush()
        db.session.expire_all()

        # call all post-commit callbacks that would have been called by the DBSessionMiddleware
        # since we are simulating a commit by flushing the session
        for callback in db.session_context.on_post_committed:
            await callback()
        db.session_context.on_post_committed.clear()

    async with AsyncClient(
        base_url="http://testserver/backend",
        event_hooks={"response": [clear_session]},
        transport=ASGITransport(app=app),
    ) as client:
        yield client



import contextlib
import importlib
import json
import logging
from dataclasses import dataclass
from functools import lru_cache
from importlib.metadata import entry_points
from pathlib import Path
from typing import TypeVar

from pydantic import BaseModel

from brickworks.core.models.base_dbmodel import BaseDBModel

logger = logging.getLogger(__name__)


def discover_modules(group: str = "brickworks.modules") -> list[str]:
    """
    Discover modules registered under the specified entry point group.
    To make modules discoverable, they should be registered in the pyproject.toml file like this:
    [project.entry-points."brickworks.modules"]
    app_mymodule = "app.mymodule"

    You only need to register local bricks in the pyproject.toml file.
    Bricks installed as packages don't need to be registered in the pyproject.toml file, because
    packages register their entry points in their own pyproject.toml file.
    """

    return [ep.value for ep in entry_points(group=group)]


@dataclass
class Module:
    path: str
    routers: list[str]
    python_module: object
    middlewares: list[str]


class ModuleJsonContent(BaseModel):
    routers: list[str] = []
    middlewares: list[str] = []
    loadme: list[str] = []


@lru_cache
def load_modules() -> list[Module]:
    """
    Load all modules from the pyproject.toml file.
    """
    module_names = discover_modules()

    modules = []
    for module in module_names:
        python_module = importlib.import_module(module)
        module_json_path = _get_module_json_path(module)
        with open(module_json_path, encoding="UTF-8") as f:
            module_json_content = ModuleJsonContent.model_validate(json.load(f))
            modules.append(
                Module(
                    path=module,
                    python_module=python_module,
                    routers=module_json_content.routers,
                    middlewares=module_json_content.middlewares,
                )
            )

            # import modules from loadme
            for loadme in module_json_content.loadme:
                importlib.import_module(loadme)
    logger.info(f"Loaded {len(modules)} modules: {[module.path for module in modules]}")
    return modules


# lru cache so we don't load models if they are already loaded
@lru_cache
def load_models() -> None:
    modules = load_modules()
    for module in modules:
        with contextlib.suppress(ModuleNotFoundError):
            importlib.import_module(module.path + ".models")


T = TypeVar("T")


def get_all_subclasses(cls: type[T]) -> set[type[T]]:
    subclasses = set(cls.__subclasses__())
    for subclass in list(subclasses):
        subclasses.update(get_all_subclasses(subclass))
    return subclasses


@lru_cache
def get_models_by_fqpn() -> dict[str, type[BaseDBModel]]:
    models = get_all_subclasses(BaseDBModel)
    # remove abstract models
    models_filtered = [model for model in models if hasattr(model, "__tablename__")]

    return {model.fqpn(): model for model in models_filtered}


def _get_module_json_path(module_name: str) -> str:
    module = importlib.import_module(module_name)
    if module.__file__ is None:
        raise ValueError(f"Module {module_name} does not have a __file__ attribute.")
    module_dir = Path(module.__file__).parent
    bricks_json_path = module_dir / "brick.json"
    return str(bricks_json_path)



import pytest
from sqlalchemy.exc import ProgrammingError

from brickworks.core import db
from brickworks.core.models.tenant_model import TenantModel
from tests.conftest import TestApp


async def test_create_delete_tenant(app: TestApp) -> None:
    """
    Test that will create a tenant and delete it again.
    """

    tenant = await TenantModel.create(
        domain="test.example.com",
        schema="test_schema",
        label="Test Tenant",
    )
    assert tenant.domain == "test.example.com"
    assert tenant.schema == "test_schema"
    assert tenant.label == "Test Tenant"

    # verify that we can fetch data from the tenant schema
    async with db(schema=tenant.schema):
        result = await TenantModel.get_list()
        assert len(result) == 0
        # if the fetch would access the master schema, it would return the tenant model itself
        # in the tenant this should be empty

    # delete the tenant
    await tenant.delete()

    # the deletion of the tenant schema is done in post commit callbacks,
    # which aren't executed in tests (because we rollback the session)
    # so we need to manually call the post commit callbacks
    for callback in db.session_context.on_post_committed:
        await callback()
    db.session_context.on_post_committed.clear()

    async with db(schema=tenant.schema):
        # verify that the tenant schema is dropped
        with pytest.raises(ProgrammingError):
            result = await TenantModel.get_list()



from authlib.integrations.base_client import OAuthError
from fastapi.responses import RedirectResponse
from httpx import AsyncClient
from pytest_mock import MockerFixture


async def test_login_redirects(client: AsyncClient, mocker: MockerFixture) -> None:
    mock_authorize_redirect = mocker.patch(
        "brickworks.core.routers.auth_router.oidc_client.authorize_redirect",
        return_value=RedirectResponse(url="https://example.com"),
    )
    response = await client.get("/api/core/auth/login", follow_redirects=False)
    assert response.status_code in (302, 307)
    assert "location" in response.headers
    mock_authorize_redirect.assert_called_once()


async def test_redirect_success(client: AsyncClient, mocker: MockerFixture) -> None:
    mock_token = {"userinfo": {"sub": "123", "name": "Test User", "email": "test@example.com"}}
    mocker.patch(
        "brickworks.core.routers.auth_router.oidc_client.authorize_access_token",
        return_value=mock_token,
    )
    response = await client.get("/api/core/auth/redirect")
    assert response.status_code in (302, 307)
    assert "location" in response.headers


async def test_redirect_oauth_error(client: AsyncClient, mocker: MockerFixture) -> None:
    mocker.patch(
        "brickworks.core.routers.auth_router.oidc_client.authorize_access_token",
        side_effect=OAuthError(error="oauth_error"),
    )
    response = await client.get("/api/core/auth/redirect", follow_redirects=False)
    assert response.status_code == 200
    assert "oauth_error" in response.text


async def test_logout_with_end_session(client: AsyncClient, mocker: MockerFixture) -> None:
    mocker.patch(
        "brickworks.core.routers.auth_router.oidc_client.load_server_metadata",
        return_value={"end_session_endpoint": "https://logout.example.com"},
    )
    response = await client.get("/api/core/auth/logout")
    assert response.status_code in (302, 307)
    assert response.headers["location"].startswith("https://logout.example.com")


async def test_logout_without_end_session(client: AsyncClient, mocker: MockerFixture) -> None:
    mocker.patch(
        "brickworks.core.routers.auth_router.oidc_client.load_server_metadata",
        return_value={},
    )
    response = await client.get("/api/core/auth/logout")
    assert response.status_code in (302, 307)
    assert "location" in response.headers


async def test_csrf_token_new(client: AsyncClient) -> None:
    response = await client.get("/api/core/auth/csrf")
    assert response.status_code == 200
    assert "csrf_token" in response.json()
    assert len(response.json()["csrf_token"]) == 64


async def test_csrf_token_existing(client: AsyncClient) -> None:
    # Use the same client context to preserve session
    response1 = await client.get("/api/core/auth/csrf")
    token1 = response1.json()["csrf_token"]
    response2 = await client.get("/api/core/auth/csrf")
    token2 = response2.json()["csrf_token"]
    assert token1 == token2



from collections.abc import Sequence
from typing import Any, TypeVar

from sqlalchemy import Select, and_

from brickworks.core import db
from brickworks.core.acl.base_policy import BasePolicy
from brickworks.core.auth.authcontext import auth_context
from brickworks.core.models.base_dbmodel import BaseDBModel
from brickworks.core.schemas.base_schema import BaseSchema
from brickworks.core.utils.sqlalchemy import TypeOrderBy, TypeWhereClause

BaseViewType = TypeVar("BaseViewType", bound="BaseView")  # pylint: disable=invalid-name


class BaseView(BaseSchema):
    """
    Base class for creating SQLAlchemy-based read-only views.

    To define a custom view, subclass BaseView and specify:
      - `__select__`: a SQLAlchemy Select statement that defines the columns and joins for the view.
      - `__policy_model_class__`: The model that should be used to apply policies.
      - `__policies__`: A list of policies to apply to the view. If not provided the models policies are used.

    If you are using joins, make sure to use the `aliased` function from SQLAlchemy to avoid name conflicts!

    Example:
        class MyCustomView(BaseView):
            field1: str
            field2: int

            __select__ = select(
                SomeModel.field1.label("field1"),
                SomeModel.field2.label("field2"),
            )
            __policy_model_class__ = SomeModel

    Usage:
        - Use `await MyCustomView.get_list(...)` to fetch multiple rows as Pydantic objects.
        - Use `await MyCustomView.get_one_or_none(...)` to fetch a single row or None.

    Filtering, ordering, and pagination are supported via method arguments.
    (same as for the database models)
    """

    __select__: Select[Any]
    __policy_model_class__: type[BaseDBModel] | None = None  # nodel class used for policies
    __policies__: list["BasePolicy"] = []  # policies to apply to the view

    @classmethod
    async def get_one_or_none_with_policies(
        cls: type[BaseViewType],
        _filter_clause: TypeWhereClause | None = None,
        **kwargs: dict[str, Any],
    ) -> BaseViewType | None:
        """
        Retrieve a single database row matching all provided key-value pairs, with custom filtering.
        This method applies all restrictive (AND) and permissive (OR) policies defined in
        __policies__ of the __model_class__.

        Filtering options:
        - Additional key-value filters can be passed as kwargs (combined with AND).
        - If _filter_clause is provided, it is added as an extra SQLAlchemy filter.

        Returns None if no match is found.
        """
        return await cls.get_one_or_none(_apply_policies=True, _filter_clause=_filter_clause, **kwargs)

    @classmethod
    async def get_one_or_none(
        cls: type[BaseViewType],
        _apply_policies: bool = False,
        _filter_clause: TypeWhereClause | None = None,
        **kwargs: dict[str, Any],
    ) -> BaseViewType | None:
        """
        Retrieve a single database row matching all provided key-value pairs, with optional policy and custom filtering.

        Filtering options:
        - _apply_policies: applies all restrictive (AND) and permissive (OR) policies defined in __policies__
        - Additional key-value filters can be passed as kwargs (combined with AND).
        - If _filter_clause is provided, it is added as an extra SQLAlchemy filter.

        Returns None if no match is found.
        """
        query = cls.__select__
        if _apply_policies:
            if not cls.__policy_model_class__:
                raise ValueError("Cannot apply policies without a model class")
            # if _apply_policies is given we apply the policies of the model class
            query = await cls.__policy_model_class__.apply_policies_to_query(
                query, user_uuid=auth_context.user_uuid, policies=cls.__policies__ or None
            )

        # filter by keys
        if kwargs:
            query = query.where(and_(*(getattr(cls, key) == value for key, value in kwargs.items())))
        # filter by custom filter clause
        if _filter_clause is not None:
            query = query.where(_filter_clause)
        result = await db.session.execute(query)
        row = result.unique().one_or_none()
        return cls(**dict(zip(row._fields, row._t, strict=True))) if row else None

    @classmethod
    async def get_list_with_policies(
        cls: type[BaseViewType],
        _filter_clause: TypeWhereClause | None = None,
        _order_by: TypeOrderBy | None = None,
        _per_page: int = -1,
        _page: int = 1,
        **kwargs: dict[str, Any],
    ) -> Sequence[BaseViewType]:
        """
        Retrieve a list of database rows with flexible filtering, ordering, and pagination.
        This method applies all restrictive (AND) and permissive (OR) policies defined in
        __policies__ of the __model_class__.

        Filtering options:
        - Additional key-value filters can be passed as kwargs (combined with AND).
        - If _filter_clause is provided, it is added as an extra SQLAlchemy filter.

        Ordering and pagination:
        - _order_by: SQLAlchemy column or list of columns to order by.
        - _per_page: If > 0, limits the number of results per page.
        - _page: Page number for paginated results (1-based).

        Returns a sequence of matching rows.
        """
        return await cls.get_list(
            _apply_policies=True,
            _filter_clause=_filter_clause,
            _order_by=_order_by,
            _per_page=_per_page,
            _page=_page,
            **kwargs,
        )

    @classmethod
    async def get_list(
        cls: type[BaseViewType],
        _apply_policies: bool = False,
        _filter_clause: TypeWhereClause | None = None,
        _order_by: TypeOrderBy | None = None,
        _per_page: int = -1,
        _page: int = 1,
        **kwargs: dict[str, Any],
    ) -> Sequence[BaseViewType]:
        """
        Retrieve a list of database rows with flexible filtering, ordering, and pagination.

        Filtering options:
        - _apply_policies: applies all restrictive (AND) and permissive (OR) policies defined in __policies__
        - Additional key-value filters can be passed as kwargs (combined with AND).
        - If _filter_clause is provided, it is added as an extra SQLAlchemy filter.

        Ordering and pagination:
        - _order_by: SQLAlchemy column or list of columns to order by.
        - _per_page: If > 0, limits the number of results per page.
        - _page: Page number for paginated results (1-based).

        Returns a sequence of matching rows.
        """
        query = cls.__select__
        if _apply_policies:
            if not cls.__policy_model_class__:
                raise ValueError("Cannot apply policies without a model class")
            # if _apply_policies is given we apply the policies of the model class
            query = await cls.__policy_model_class__.apply_policies_to_query(
                query, user_uuid=auth_context.user_uuid, policies=cls.__policies__ or None
            )

        # filter by keys
        if kwargs:
            query = query.where(and_(*(getattr(cls, key) == value for key, value in kwargs.items())))
        # filter by custom filter clause
        if _filter_clause is not None:
            query = query.where(_filter_clause)

        if _order_by is not None:
            query = query.order_by(*_order_by if isinstance(_order_by, list) else [_order_by])
        if _per_page > 0:
            query = query.limit(_per_page).offset((_page - 1) * _per_page)

        result = await db.session.execute(query)
        rows = result.unique()
        return [cls(**dict(zip(row._fields, row._t, strict=True))) for row in rows]



import logging
import sys

import typer
import uvicorn

from brickworks.core.constants import BASE_DIR
from brickworks.core.server import create_app
from brickworks.mason.brick_app import brick_app
from brickworks.mason.db_app import db_app

logging.basicConfig(level=logging.INFO)

app = typer.Typer()

sys.path.append(str(BASE_DIR))

app.add_typer(db_app, name="db", help="Database commands")
app.add_typer(brick_app, name="brick", help="Brick commands")


@app.command()
def runserver(host: str = "0.0.0.0", port: int = 8000) -> None:  # nosec
    server = create_app()
    uvicorn.run(server, host=host, port=port)


def main() -> None:
    app()


if __name__ == "__main__":
    main()



from .base_dbmodel import BaseDBModel
from .role_model import RoleModel
from .tenant_model import TenantModel
from .user_model import UserModel

__all__ = ["BaseDBModel", "UserModel", "RoleModel", "TenantModel"]



import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING

from brickworks.core.utils.sqlalchemy import TypeWhereClause

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from brickworks.core.models.base_dbmodel import BaseDBModel


class PolicyTypeEnum(Enum):
    """
    Enum to distinguish between restrictive and permissive policies.

    - RESTRICTIVE: All restrictive policies must be satisfied (AND logic).
    - PERMISSIVE: At least one permissive policy must be satisfied (OR logic).
    """

    RESTRICTIVE = "restrictive"
    PERMISSIVE = "permissive"


class BasePolicy(ABC):
    """
    Abstract base class for access control policies.

    Policies are used to restrict or permit access to database rows at the query level.
    Each policy must define its type (restrictive or permissive) and implement the __call__ method,
    which returns a SQLAlchemy filter clause to be applied to a query.

    - Restrictive policies are combined with AND: they prevent access to any row
    that doesn't fulfill their condition.
    - Permissive policies are combined with OR: they add access to rows that fulfill their condition.

    The __call__ method receives the user UUID and the model class, and should return a FilterClause
    suitable for use in a WHERE clause.
    """

    policy_type: PolicyTypeEnum

    @abstractmethod
    async def filter(self, user_uuid: str | None, obj_class: type["BaseDBModel"]) -> TypeWhereClause:
        """
        Return a SQLAlchemy filter clause for this policy.

        Args:
            user_uuid: The UUID of the user for whom access is being checked. If None, assume public access.
            obj_class: The SQLAlchemy model class being queried.

        Returns:
            ClauseElement: A SQLAlchemy expression to be used in a WHERE clause.
        """
        raise NotImplementedError("Policy must implement filter method")



from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKey, String, Table, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship

from brickworks.core.models.base_dbmodel import UUID_LENGTH, Base, BaseDBModel

if TYPE_CHECKING:
    from brickworks.core.models.user_model import UserModel

user_role_table = Table(
    "user_role_table",
    Base.metadata,
    Column("user_uuid", ForeignKey("core_users.uuid"), index=True),
    Column("role_uuid", ForeignKey("core_roles.uuid")),
)

role_acl_table = Table(
    "role_acl_table",
    Base.metadata,
    Column("role_name", ForeignKey("core_roles.role_name"), index=True),
    Column("object_uuid", String(UUID_LENGTH), index=True),
    Column("object_fqpn", String(255)),
    Column("permission", String(64)),
    UniqueConstraint("role_name", "object_uuid", "permission", name="uq_role_acl_all"),
)


class RoleModel(BaseDBModel):
    __tablename__ = "core_roles"
    role_name: Mapped[str] = mapped_column(String(64), nullable=False, unique=True)
    users: Mapped[list["UserModel"]] = relationship(
        secondary="user_role_table", back_populates="roles", lazy="noload", init=False
    )



import logging
import secrets

from authlib.integrations.base_client import OAuthError
from fastapi import APIRouter, Response
from fastapi.responses import HTMLResponse, RedirectResponse
from pydantic import ValidationError
from starlette.requests import Request

from brickworks.core.auth.oidc import OIDCUserInfo, get_or_create_user_from_userinfo, oidc_client
from brickworks.core.schemas.auth_schemas import CSRFTokenResponse
from brickworks.core.settings import settings

logger = logging.getLogger(__name__)
auth_router = r = APIRouter(prefix="/auth")


@r.get("/login")
async def _login(request: Request) -> RedirectResponse:
    """
    Will redirect the user to the OIDC provider for authentication.
    """
    redirect_uri = request.url_for("_redirect")
    # we need to help out mypy here, because authlib isn't typed
    response: RedirectResponse = await oidc_client.authorize_redirect(request, redirect_uri)
    return response


@r.get("/redirect")
async def _redirect(request: Request) -> Response:
    """
    This is the redirect URL that the OIDC provider will call after the user has logged in.
    If the login was successful we set the user in the session and redirect to the POST_LOGIN_REDIRECT_URI.
    """
    try:
        token = await oidc_client.authorize_access_token(request)
    except OAuthError as error:
        # TODO: redirect to error page?
        return HTMLResponse(f"<h1>{error.error}</h1>")
    userinfo = token.get("userinfo")
    if not userinfo:
        return HTMLResponse("<h1>Userinfo not found</h1>")
    try:
        userinfo_model = OIDCUserInfo.model_validate(userinfo)
    except ValidationError as error:
        logger.error("Userinfo validation error!", exc_info=error)
        return HTMLResponse(f"<h1>{error.errors()}</h1>")
    user = await get_or_create_user_from_userinfo(userinfo_model)
    request.session["user_uuid"] = user.uuid

    return RedirectResponse(url=settings.POST_LOGIN_REDIRECT_URI)


@r.get("/logout")
async def _logout(request: Request) -> RedirectResponse:
    """
    Will logout the user by clearing the session.
    If the OIDC provider supports it, we will also redirect to the logout URL of the OIDC provider.
    After that we redirect to the POST_LOGOUT_REDIRECT_URI.
    """
    request.session.pop("user", None)
    metadata = await oidc_client.load_server_metadata()
    if "end_session_endpoint" in metadata:
        logout_url = f"{metadata['end_session_endpoint']}?post_logout_redirect_uri={settings.POST_LOGOUT_REDIRECT_URI}"
        return RedirectResponse(url=logout_url)
    return RedirectResponse(url=settings.POST_LOGOUT_REDIRECT_URI)


@r.get("/csrf")
async def _csrf(request: Request) -> CSRFTokenResponse:
    """
    This endpoint returns a CSRF token for the user.
    """
    token = request.session.get("csrf_token")
    if not token:
        token = secrets.token_hex(32)
        request.session["csrf_token"] = token
    return CSRFTokenResponse(csrf_token=token)



import os
from pathlib import Path

import tomli
import tomli_w
import typer

from brickworks.core.constants import BASE_DIR
from brickworks.core.module_loader import ModuleJsonContent

brick_app = typer.Typer()


def _create_python_module(path: str) -> None:
    """
    Create a Python module by creating an empty __init__.py file.
    """
    p = Path(path)
    p.mkdir(parents=True, exist_ok=True)
    (p / "__init__.py").touch()


def _add_brick_to_toml(name: str, namespace: str) -> None:
    """
    Add the brick to the pyproject.toml file under entry points.
    """
    toml_path = os.path.join(BASE_DIR, "pyproject.toml")
    with open(toml_path, "rb") as f:
        config = tomli.load(f)

    # Ensure the necessary sections exist
    config.setdefault("project", {}).setdefault("entry-points", {}).setdefault("brickworks.modules", {})
    config.setdefault("tool", {}).setdefault("setuptools", {}).setdefault("packages", {}).setdefault(
        "find", {}
    ).setdefault("include", [])

    # Add the new module to the entry points, so that we can discovery it as a brick
    entry_points = config["project"]["entry-points"]["brickworks.modules"]
    entry_points[f"{namespace}_{name}"] = f"{namespace}.{name}"

    # Ensure the module is included in the setuptools packages find section
    include_list = config["tool"]["setuptools"]["packages"]["find"]["include"]
    include_list.append(f"{namespace}.{name}")

    # Write the updated data back to the TOML file
    with open(toml_path, "wb") as file:
        file.write(tomli_w.dumps(config).encode())


def _create_namespace_package(namespace: str) -> None:
    """
    Create a namespace package for the given namespace.
    Does nothing if the namespace already exists.
    """
    namespace_dir = os.path.join(BASE_DIR, namespace)
    if not os.path.exists(namespace_dir):
        os.makedirs(namespace_dir)
    if not os.path.exists(os.path.join(namespace_dir, "__init__.py")):
        with open(os.path.join(namespace_dir, "__init__.py"), "w"):
            pass


@brick_app.command()
def create(name: str, namespace: str = "app") -> None:
    """Create a new brick."""
    typer.echo(f"Creating brick: {namespace}.{name}")

    # make sure the namespace packages folder exists
    _create_namespace_package(namespace)

    # Create the brick directory
    brick_dir = f"{BASE_DIR}/{namespace}/{name}"
    try:
        _create_python_module(brick_dir)
    except FileExistsError:
        typer.echo(f"Path {brick_dir} already exists")
        return

    # create the brick.json file
    module_json_content = ModuleJsonContent()
    with open(os.path.join(brick_dir, "brick.json"), "w", encoding="UTF-8") as f:
        f.write(module_json_content.model_dump_json(indent=4))

    # create the models, schemas and routers modules
    for mod in ("models", "schemas", "routers"):
        _create_python_module(os.path.join(brick_dir, mod))

    # add the brick to the pyproject.toml file
    _add_brick_to_toml(name, namespace)
    typer.echo(f"Brick {namespace}.{name} created successfully.")
    typer.echo("Don't forget to install the project with `pip install -e .`")



import logging
from datetime import timedelta

import jwt  # pyjwt
from fastapi import Request
from pydantic import BaseModel

from brickworks.core.exceptions import UnauthorizedException
from brickworks.core.settings import settings
from brickworks.core.utils.timeutils import now_utc

logger = logging.getLogger(__name__)


class JWTPayload(BaseModel):
    sub: str
    scope: list[str]
    iat: int
    exp: int


async def generate_jwt(scope: list[str], user_uuid: str, ttl: timedelta = timedelta(hours=1)) -> str:
    secret = settings.JWT_SECRET
    now = now_utc()
    exp = now + ttl
    payload = {
        "sub": user_uuid,
        "scope": scope,
        "iat": int(now.timestamp()),
        "exp": int(exp.timestamp()),
    }
    token = jwt.encode(payload, secret, algorithm="HS256")
    return token.decode("utf8")


def validate_jwt(token: str, required_scopes: list[str] | None = None) -> JWTPayload:
    """
    Validates and decodes a JWT. Raises jwt exceptions if invalid or expired.
    If required_scopes is provided, ensures all are present in the token's scope claim.
    Returns the decoded payload as a JWTPayload model.
    """
    secret = settings.JWT_SECRET
    payload = jwt.decode(token, secret, algorithms=["HS256"])
    jwt_payload = JWTPayload.model_validate(payload)
    if required_scopes:
        missing = [scope for scope in required_scopes if scope not in jwt_payload.scope]
        if missing:
            logger.error("Missing required scopes: %s", ",".join(missing))
            raise UnauthorizedException
    return jwt_payload


def get_jwt_payload(
    request: Request,
    required_scopes: list[str] | None = None,
) -> JWTPayload:
    """
    FastAPI dependency to extract and validate JWT from the Authorization header.
    Raises HTTPException if missing or invalid.
    """
    auth_header = request.headers.get("Authorization")
    if not auth_header or not auth_header.startswith("Bearer "):
        raise UnauthorizedException("Missing token")
    token = auth_header.removeprefix("Bearer ").strip()
    try:
        return validate_jwt(token, required_scopes)
    except Exception as e:
        logger.error("Token auth failed", exc_info=e)
        raise UnauthorizedException("Token validation failed") from e



import urllib.parse
from contextvars import ContextVar
from types import TracebackType
from typing import Optional

from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response

from brickworks.core.settings import settings

_auth_context_var: ContextVar[Optional["AuthContext"]] = ContextVar("auth_context", default=None)


def _get_context_var() -> "AuthContext":
    context = _auth_context_var.get()
    if context is None:
        raise RuntimeError("AuthContext not set")
    return context


class AuthContextMeta(type):
    @property
    def user_uuid(cls) -> str | None:
        return _get_context_var().user_uuid

    @property
    def tenant_schema(cls) -> str | None:
        return _get_context_var().tenant_schema


class AuthContext(metaclass=AuthContextMeta):
    def __init__(self, user_uuid: str | None = None, tenant_schema: str | None = None) -> None:
        self.user_uuid = user_uuid
        self.tenant_schema = tenant_schema or settings.MASTER_DB_SCHEMA
        self.token = _auth_context_var.set(self)

    async def __aenter__(self) -> type["AuthContext"]:
        return type(self)

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        _auth_context_var.reset(self.token)


class AuthContextMiddleware(BaseHTTPMiddleware):
    async def _extract_tenant_from_request(self, request: Request) -> str | None:
        """
        Extract tenant schema from the request.
        Get the host domain and map it to the tenant schema.
        If the request is not for a tenant, return the master schema.
        """
        host_header = request.headers.get("host", "")
        # Prepend '//' so urlparse treats host_header as a netloc (hostname:port)
        parsed = urllib.parse.urlparse(f"//{host_header}")
        domain = parsed.hostname or ""
        if not domain:
            return None

        from brickworks.core.models.tenant_model import get_domain_schema_mapping

        mapping = await get_domain_schema_mapping()
        return mapping.get(domain)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        if settings.MULTI_TENANCY_ENABLED:
            tenant_schema = await self._extract_tenant_from_request(request)
        else:
            tenant_schema = settings.MASTER_DB_SCHEMA
        if not tenant_schema:
            return Response(
                "Tenant not found",
                status_code=404,
                headers={"Content-Type": "text/plain"},
            )
        async with auth_context(
            user_uuid=request.session.get("user_uuid"), tenant_schema=await self._extract_tenant_from_request(request)
        ):
            response = await call_next(request)
        return response


auth_context = AuthContext



from unittest.mock import patch

from brickworks.core.acl.policies import (
    AllowActiveUserAccessPolicy,
    AllowPublicAccessPolicy,
    RoleAllowPolicy,
    RoleBasedAccessPolicy,
)
from brickworks.core.auth.authcontext import AuthContext
from brickworks.core.models.role_model import RoleModel
from brickworks.core.models.user_model import UserModel, UserStatusEnum
from tests.conftest import TestApp
from tests.core.utils import create_test_user


async def test_get_list(app: TestApp) -> None:
    # Add test users
    user_alice = await create_test_user("Alice", "Smith")
    user_bob = await create_test_user("Bob", "Smith")
    result = await UserModel.get_list()

    assert len(result) >= 2
    assert user_alice in result
    assert user_bob in result


async def test_no_policy_set(app: TestApp) -> None:
    user_alice = await create_test_user("Alice", "Smith")
    await create_test_user("Bob", "Smith")

    async with AuthContext(user_alice.uuid):
        result = await UserModel.get_list_with_policies()
        # No policies are applied so we should see nobody
        assert len(result) == 0


async def test_allow_public_access_policy(app: TestApp) -> None:
    await create_test_user("Alice", "Smith")
    await create_test_user("Bob", "Smith")

    async with AuthContext(None):
        # no logged in user
        result = await UserModel.get_list_with_policies()
        # No policies are applied so we should see nobody
        assert len(result) >= 0

    # add AllowPublicAccessPolicy to the UserModel
    with patch.object(UserModel, "__policies__", [AllowPublicAccessPolicy()]):
        async with AuthContext(None):
            # no logged in user
            result = await UserModel.get_list_with_policies()
            # all users should be visible
            assert len(result) >= 2


async def test_allow_active_user_access_policy(app: TestApp) -> None:
    user_alice = await create_test_user("Alice", "Smith")
    user_bob = await create_test_user("Bob", "Smith")
    user_bob.status = UserStatusEnum.INACTIVE
    await user_bob.persist()

    # add AllowActiveUserAccessPolicy to the UserModel
    with patch.object(UserModel, "__policies__", [AllowActiveUserAccessPolicy()]):
        async with AuthContext(user_alice.uuid):
            result = await UserModel.get_list_with_policies()
            # Alice is active, so she should see Bob
            assert len(result) >= 2

        async with AuthContext(user_bob.uuid):
            result = await UserModel.get_list_with_policies()
            # Bob is inactive, so he should not see anyone
            assert len(result) == 0

        async with AuthContext(None):
            result = await UserModel.get_list_with_policies()
            # no logged in user
            assert len(result) == 0


async def test_role_allow_policy(app: TestApp) -> None:
    role1 = await RoleModel(role_name="role1").persist()
    user_alice = await create_test_user("Alice", "Smith")
    user_bob = await create_test_user("Bob", "Smith")

    await user_bob.give_role_permission("role1", "read")

    # add RoleAllowPolicy to the UserModel
    with patch.object(UserModel, "__policies__", [RoleAllowPolicy(role1.role_name, "read")]):
        # query as Alice
        async with AuthContext(user_alice.uuid):
            result = await UserModel.get_list_with_policies()
            # Alice does not have the role, so she should not see Bob
            assert len(result) == 0

            # give Alice the role
            await user_alice.add_role(role1)
            result = await UserModel.get_list_with_policies()
            # Alice has the role, so she should see Bob
            assert len(result) == 1


async def test_role_based_access_policy_permissive(app: TestApp) -> None:
    role1 = await RoleModel(role_name="role1").persist()
    user_alice = await create_test_user("Alice", "Smith")
    await create_test_user("Bob", "Smith")

    # add RoleBasedAccessPolicy to the UserModel
    with patch.object(UserModel, "__policies__", [RoleBasedAccessPolicy(role1.role_name)]):
        # query as Alice
        async with AuthContext(user_alice.uuid):
            result = await UserModel.get_list_with_policies()

            # Alice does not have the role, so she should not see Bob
            assert len(result) == 0

            # give Alice the role
            await user_alice.add_role(role1)
            result = await UserModel.get_list_with_policies()
            # Alice has the role, so she should see all users
            assert len(result) >= 2


async def test_role_based_access_policy_restrictive(app: TestApp) -> None:
    role1 = await RoleModel(role_name="role1").persist()
    user_alice = await create_test_user("Alice", "Smith")
    await create_test_user("Bob", "Smith")

    # add RoleBasedAccessPolicy to the UserModel set to retrictive
    with patch.object(UserModel, "__policies__", [RoleBasedAccessPolicy(role1.role_name, restrictive=True)]):
        # query as Alice
        async with AuthContext(user_alice.uuid):
            result = await UserModel.get_list_with_policies()

            # Alice does not have the role, so she should not see Bob
            assert len(result) == 0

            # give Alice the role
            await user_alice.add_role(role1)
            result = await UserModel.get_list_with_policies()

            # Alice still can't see anyone because the policy is restrictive
            assert len(result) == 0

    # also add AllowActiveUserAccessPolicy to the UserModel
    with patch.object(
        UserModel,
        "__policies__",
        [RoleBasedAccessPolicy(role1.role_name, restrictive=True), AllowActiveUserAccessPolicy()],
    ):
        # query as Alice
        async with AuthContext(user_alice.uuid):
            # remove role from Alice
            await user_alice.remove_role(role1)

            # Alice does not have the role, so she should not see Bob, dispite the AllowActiveUserAccessPolicy
            result = await UserModel.get_list_with_policies()
            assert len(result) == 0

            # give Alice the role
            await user_alice.add_role(role1)
            result = await UserModel.get_list_with_policies()
            # Alice has the role, so she should see all users
            assert len(result) >= 2



from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    SESSION_SECRET: str = ""  # used for signing session data
    JWT_SECRET: str = ""  # used for signing jwts

    MULTI_TENANCY_ENABLED: bool = False
    MASTER_DOMAIN: str = "example.com"
    MASTER_DB_SCHEMA: str = "public"

    # Database settings
    DB_HOST: str = "127.0.0.1:5432"
    DB_USER: str = "postgres"
    DB_PASSWORD: str = "postgres"
    DB_NAME: str = "postgres"

    USE_REDIS: bool = True
    REDIS_HOST: str = "localhost"
    REDIS_PORT: int = 6379
    REDIS_DB: int = 0
    REDIS_PASSWORD: str | None = None

    # OIDC
    OIDC_CLIENT_ID: str = ""
    OIDC_CLIENT_SECRET: str = ""
    OIDC_DISCOVERY_URL: str = "https://accounts.google.com/.well-known/openid-configuration"
    OIDC_PROVIDER_NAME: str = "google"
    POST_LOGOUT_REDIRECT_URI: str = "/"
    POST_LOGIN_REDIRECT_URI: str = "/"


settings = Settings()



from datetime import datetime

from pydantic import BaseModel, ConfigDict

from brickworks.core.utils.timeutils import convert_datetime_to_iso_8601_with_z_suffix


class BaseSchema(BaseModel):
    """
    Base schema for all schemas in the project.
    It converts datetime objects to ISO 8601 format with Z suffix.
    Make sure that your datetime objects are timezone unaware!
    """

    model_config = ConfigDict(
        json_encoders={datetime: convert_datetime_to_iso_8601_with_z_suffix}, from_attributes=True
    )



import json
from base64 import b64decode, b64encode
from typing import Any
from uuid import uuid4

import itsdangerous
from fastapi import Request
from itsdangerous.exc import BadTimeSignature, SignatureExpired
from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from brickworks.core.cache import cache
from brickworks.core.settings import settings


async def get_session(request: Request) -> dict[str, Any]:
    """
    Can be used as a dependency to get the session data.
    """
    return request.session


class SessionMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        cookie_name: str,
        max_age: int = 14 * 24 * 60 * 60,  # 14 days, in seconds
        same_site: str = "lax",
        https_only: bool = False,
        domain: str | None = None,
    ) -> None:
        """Session Middleware, that stores the session data in the cache.
        This allows us to invalidate sessions by deleting the session data from the cache.
        The session id is stored in a cookie, which is signed with a secret key, to prevent tampering.

        Args:
            app: The ASGIApp
            cookie_name: The name of the cookie used to store the session id.
            max_age: The Max-Age of the cookie (Default to 14 days).
            same_site: The SameSite attribute of the cookie (Defaults to lax).
            https_only: Whether to make the cookie https only (Defaults to False).
            domain: The domain associated to the cookie (Default to None).
        """
        self.app = app
        self.cache_namespace = "session"
        self.signer = itsdangerous.TimestampSigner(settings.SESSION_SECRET)
        self.cookie_name = cookie_name
        self.max_age = max_age
        self.domain = domain

        self._cookie_session_id_field = "_cssid"

        self.security_flags = f"httponly; samesite={same_site}"
        if https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += "; secure"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        initial_session_was_empty = True

        # Always use the same session id if present, otherwise generate a new one
        if self.cookie_name in connection.cookies:
            data = connection.cookies[self.cookie_name].encode("utf-8")
            try:
                data = self.signer.unsign(data, max_age=self.max_age)
                session_key = json.loads(b64decode(data)).get(self._cookie_session_id_field)
                if not isinstance(session_key, str) or not session_key:
                    session_key = str(uuid4())
                scope["session"] = json.loads(await cache.get_key(session_key, self.cache_namespace) or "{}")
                scope["__session_key"] = session_key
                if scope["session"] is None:
                    scope["session"] = {}

                initial_session_was_empty = False
            except (BadTimeSignature, SignatureExpired, json.JSONDecodeError):
                session_key = str(uuid4())
                scope["session"] = {}
                scope["__session_key"] = session_key
        else:
            session_key = str(uuid4())
            scope["session"] = {}
            scope["__session_key"] = session_key

        async def send_wrapper(message: Message, **_: Any) -> None:  # noqa: ANN401
            if message["type"] == "http.response.start":
                session_key = scope.get("__session_key")
                if not isinstance(session_key, str) or not session_key:
                    session_key = str(uuid4())

                if scope["session"]:
                    await cache.set_key(
                        session_key, json.dumps(scope["session"]), namespace=self.cache_namespace, expire=self.max_age
                    )
                    cookie_data = {self._cookie_session_id_field: session_key}

                    data = b64encode(json.dumps(cookie_data).encode("utf-8"))
                    data = self.signer.sign(data)

                    headers = MutableHeaders(scope=message)
                    header_value = self._construct_cookie(data=data)
                    headers.append("Set-Cookie", header_value)

                elif not initial_session_was_empty:
                    await cache.delete_key(session_key, namespace=self.cache_namespace)

                    headers = MutableHeaders(scope=message)
                    header_value = self._construct_cookie(data=None)
                    headers.append("Set-Cookie", header_value)

                else:
                    # Always refresh the cookie expiry even if session is empty
                    cookie_data = {self._cookie_session_id_field: session_key}
                    data = b64encode(json.dumps(cookie_data).encode("utf-8"))
                    data = self.signer.sign(data)
                    headers = MutableHeaders(scope=message)
                    header_value = self._construct_cookie(data=data)
                    headers.append("Set-Cookie", header_value)

            await send(message)

        await self.app(scope, receive, send_wrapper)

    def _construct_cookie(self, data: bytes | None = None) -> str:
        if data is None:
            # If the session is empty, we need to delete the cookie
            cookie = (
                f"{self.cookie_name}=null; Path=/; Expires="
                f"Thu, 01 Jan 1970 00:00:00 GMT; Max-Age=0; {self.security_flags}"
            )
        else:
            cookie = f"{self.cookie_name}={data.decode('utf-8')}; Path=/; Max-Age={self.max_age}; {self.security_flags}"
        if self.domain:
            cookie = f"{cookie}; Domain={self.domain}"
        return cookie



from enum import Enum

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from brickworks.core.exceptions import NotFoundException
from brickworks.core.models.base_dbmodel import BaseDBModel
from brickworks.core.models.role_model import RoleModel


class UserStatusEnum(str, Enum):
    """
    Enum for user status.
    """

    ACTIVE = "active"
    INACTIVE = "inactive"
    BANNED = "banned"
    DELETED = "deleted"


class UserModel(BaseDBModel):
    """
    User model representing a user in the system.

    To access the users roles use: await awaitable_attrs.roles
    """

    __tablename__ = "core_users"

    sub: Mapped[str] = mapped_column(String(255), unique=True, nullable=False)  # OpenID Connect subject identifier
    name: Mapped[str] = mapped_column(String(255), nullable=False)  # full name e.g. "John Doe"
    given_name: Mapped[str] = mapped_column(String(255), nullable=False)
    family_name: Mapped[str] = mapped_column(String(255), nullable=False)
    email: Mapped[str] = mapped_column(String(255), unique=True, nullable=False)
    email_verified: Mapped[bool] = mapped_column(nullable=False, default=False)
    phone_number: Mapped[str] = mapped_column(String(255), nullable=False, default="")
    phone_number_verified: Mapped[bool] = mapped_column(nullable=False, default=False)
    locale: Mapped[str] = mapped_column(String(10), nullable=False, default="en")
    status: Mapped[UserStatusEnum] = mapped_column(String(32), nullable=False, default=UserStatusEnum.ACTIVE)

    roles: Mapped[list["RoleModel"]] = relationship(
        "RoleModel", back_populates="users", secondary="user_role_table", lazy="noload", init=False
    )

    async def add_role(self, role: str | RoleModel) -> None:
        """
        Add a role to the user. Accepts a role name (str) or RoleModel instance.
        Persists the change to the database.
        Raises NotFoundException if the role name does not exist.
        """
        if isinstance(role, str):
            # assume its a role name and fetch it
            role_ = await RoleModel.get_one_or_none(role_name=role)
            if not role_:
                raise NotFoundException(f"Could not find role {role}")
        else:
            role_ = role

        if role_ not in await self.awaitable_attrs.roles:
            self.roles.append(role_)

        await self.persist()

    async def remove_role(self, role: str | RoleModel) -> None:
        """
        Remove a role from the user. Accepts a role name (str) or RoleModel instance.
        Persists the change to the database.
        """
        if isinstance(role, str):
            role_: RoleModel | None = next((r for r in await self.awaitable_attrs.roles if r.role_name == role), None)
            if role_:
                self.roles.remove(role_)
        else:
            if role in await self.awaitable_attrs.roles:
                self.roles.remove(role)
        await self.persist()

    async def has_role(self, role: str | RoleModel) -> bool:
        """
        Check if the user has a given role. Accepts a role name (str) or RoleModel instance.
        Returns True if the user has the role, False otherwise.
        """
        if isinstance(role, str):
            role_ = next((r for r in await self.awaitable_attrs.roles if r.role_name == role), None)
            return role_ is not None
        else:
            return role in await self.awaitable_attrs.roles



from collections.abc import Awaitable, Callable

from fastapi import Request
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response

SAFE_METHODS = ("GET", "HEAD", "OPTIONS")


class CSRFMiddleware(BaseHTTPMiddleware):
    """
    CSRF Middleware to protect against Cross-Site Request Forgery attacks.
    On unsafe methods (POST, PUT, DELETE) it checks if the CSRF token in the request header matches the CSRF token
    in the session.

    Needs to be added BEFORE the SessionMiddleware, because it needs the session to be available.
    (middleware is executed in the reverse order it is added to the app)
    """

    async def dispatch(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        if request.method in SAFE_METHODS:
            # no CSRF token is required for safe methods
            return await call_next(request)

        session_csrf_token = request.session.get("csrf_token")
        request_csrf_token = request.headers.get("X-CSRF-Token")

        if not request_csrf_token:
            return JSONResponse(
                content={
                    "detail": [
                        {
                            "loc": ["header", "X-CSRF-Token"],
                            "msg": "CSRF token is missing.",
                            "type": "csrf_token_missing",
                        }
                    ]
                },
                status_code=403,
            )
        elif request_csrf_token != session_csrf_token:
            return JSONResponse(
                content={
                    "detail": [
                        {
                            "loc": ["header", "X-CSRF-Token"],
                            "msg": "CSRF token mismatch.",
                            "type": "csrf_token_mismatch",
                        }
                    ]
                },
                status_code=403,
            )
        else:
            # CSRF token is valid, proceed with the request
            return await call_next(request)



import asyncio
import logging
from collections.abc import Awaitable, Callable
from typing import TypeVar

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class BaseSignal(BaseModel):
    """
    Base class for signals. All signals should inherit from this class.
    """

    pass


C = TypeVar("C", bound=BaseSignal)


class Signals:
    """
    Signal system for connecting and emitting events with type-safe signal classes.

    Usage:
        @signal.connect(MySignal)
        def my_listener(sig: MySignal):
            ...
        await signal.emit(MySignal(...))
    """

    def __init__(self) -> None:
        """
        Initialize the Signals system.
        """
        self.listeners: dict[type[BaseSignal], list[Callable[..., None | Awaitable[None]]]] = {}

    def connect(
        self, signal_type: type[C]
    ) -> Callable[[Callable[[C], None | Awaitable[None]]], Callable[[C], None | Awaitable[None]]]:
        """
        Register a listener for a specific signal type.
        The listener function must accept a single argument of the signal type.
        Usage:
            @signal.connect(MySignal)
            def on_my_signal(sig: MySignal):
                ...
        """

        def decorator(listener: Callable[[C], None | Awaitable[None]]) -> Callable[[C], None | Awaitable[None]]:
            self.listeners.setdefault(signal_type, []).append(listener)
            return listener

        return decorator

    async def emit(self, sig: BaseSignal) -> None:
        """
        Emit a signal instance, calling all registered listeners.
        Usage:
            await signal.emit(MySignal(...))
        """
        signal_type = type(sig)
        for listener in self.listeners.get(signal_type, []):
            if asyncio.iscoroutinefunction(listener):
                await listener(sig)
            else:
                listener(sig)

    def disconnect(self, listener: Callable[[C], None | Awaitable[None]]) -> None:
        """
        Disconnect a previously registered listener from all signal types.
        Usage:
            signal.disconnect(my_listener)
        """
        for listeners in self.listeners.values():
            if listener in listeners:
                listeners.remove(listener)

    async def emit_no_except(self, sig: BaseSignal) -> None:
        """
        Emit a signal instance, calling all registered listeners.
        Exceptions in listeners are caught and logged.
        Usage:
            await signal.emit_no_except(MySignal(...))
        """
        signal_type = type(sig)
        for listener in self.listeners.get(signal_type, []):
            if asyncio.iscoroutinefunction(listener):
                try:
                    await listener(sig)
                except Exception as e:
                    logger.error(f"Error firing signal {signal_type}: {e}")
            else:
                try:
                    listener(sig)
                except Exception as e:
                    logger.error(f"Error firing signal {signal_type}: {e}")


signals: Signals = Signals()



import pytest

from brickworks.core.signals import BaseSignal, Signals


class TestSignal(BaseSignal):
    value: int


@pytest.mark.asyncio
async def test_sync_listener_called() -> None:
    signals: Signals = Signals()
    called = {}

    @signals.connect(TestSignal)
    def listener(sig: TestSignal) -> None:
        called["value"] = sig.value

    sig = TestSignal(value=42)
    await signals.emit(sig)
    assert called["value"] == 42


@pytest.mark.asyncio
async def test_async_listener_called() -> None:
    signals: Signals = Signals()
    called = {}

    @signals.connect(TestSignal)
    async def listener(sig: TestSignal) -> None:
        called["value"] = sig.value

    sig = TestSignal(value=99)
    await signals.emit(sig)
    assert called["value"] == 99


@pytest.mark.asyncio
async def test_disconnect() -> None:
    signals: Signals = Signals()
    called = {}

    @signals.connect(TestSignal)
    def listener(sig: TestSignal) -> None:
        called["value"] = sig.value

    signals.disconnect(listener)
    sig = TestSignal(value=123)
    await signals.emit(sig)
    assert "value" not in called


@pytest.mark.asyncio
async def test_emit_no_except() -> None:
    signals: Signals = Signals()
    called = {}

    @signals.connect(TestSignal)
    def listener(sig: TestSignal) -> None:
        called["value"] = sig.value
        raise ValueError("fail")

    # Should not raise
    sig = TestSignal(value=1)
    await signals.emit_no_except(sig)
    assert "value" in called



from collections.abc import Generator
from unittest.mock import patch

import pytest

from brickworks.core.auth.authcontext import AuthContext
from brickworks.core.cache import cache
from brickworks.core.settings import settings
from tests.conftest import TestApp

TCacheBackendFixture = Generator[None, None, None]


@pytest.fixture(params=[False, True], ids=["memory", "redis"])
def cache_backend(request: pytest.FixtureRequest) -> Generator[None, None, None]:
    with patch.object(settings, "USE_REDIS", request.param):
        yield


@pytest.mark.asyncio
async def test_set_and_get(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    await cache.set_key("foo", "bar")
    value = await cache.get_key("foo")
    assert value == "bar"


async def test_set_and_get_with_tenant(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    # set everything with expire=2, to automatically clean up after the test

    # set in master tenant (tests run in master tenant by default)
    await cache.set_key("foo", "bar", namespace="test", expire=2)
    async with AuthContext(tenant_schema="test_schema"):
        assert await cache.get_key("foo", namespace="test") is None  # key doesn't exist in test tenant
        await cache.set_key("foo", "bar_test", namespace="test", expire=2)  # set in test tenant

        assert await cache.get_key("foo", namespace="test", master_tenant=True) == "bar"  # from master tenant
        assert await cache.get_key("foo", namespace="test") == "bar_test"  # from test tenant


@pytest.mark.asyncio
async def test_delete(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    await cache.set_key("delkey", "todelete")
    await cache.delete_key("delkey")
    value = await cache.get_key("delkey")
    assert value is None


@pytest.mark.asyncio
async def test_delete_with_tenant(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    # Set in master tenant
    await cache.set_key("delkey", "master_value", namespace="test", expire=2)
    async with AuthContext(tenant_schema="test_schema"):
        # Should not see master tenant's key
        assert await cache.get_key("delkey", namespace="test") is None
        # Set in tenant
        await cache.set_key("delkey", "tenant_value", namespace="test", expire=2)
        # Delete in tenant
        await cache.delete_key("delkey", namespace="test")
        assert await cache.get_key("delkey", namespace="test") is None
        # Master tenant's value should still exist
        assert await cache.get_key("delkey", namespace="test", master_tenant=True) == "master_value"
    # Delete in master
    await cache.delete_key("delkey", namespace="test")
    assert await cache.get_key("delkey", namespace="test") is None


@pytest.mark.asyncio
async def test_push_and_pop_queue(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    await cache.push_to_queue("q", "a")
    await cache.push_to_queue("q", "b")
    v1 = await cache.pop_from_queue("q")
    v2 = await cache.pop_from_queue("q")
    v3 = await cache.pop_from_queue("q")
    assert v1 == "a"
    assert v2 == "b"
    assert v3 is None


@pytest.mark.asyncio
async def test_queue_with_tenant(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    # Push to queue in master tenant
    await cache.push_to_queue("q", "master_a", namespace="test")
    async with AuthContext(tenant_schema="test_schema"):
        # Queue should be empty in tenant
        assert await cache.pop_from_queue("q", namespace="test") is None
        # Push to queue in tenant
        await cache.push_to_queue("q", "tenant_a", namespace="test")
        await cache.push_to_queue("q", "tenant_b", namespace="test")
        v1 = await cache.pop_from_queue("q", namespace="test")
        v2 = await cache.pop_from_queue("q", namespace="test")
        v3 = await cache.pop_from_queue("q", namespace="test")
        assert v1 == "tenant_a"
        assert v2 == "tenant_b"
        assert v3 is None
    # Master queue should still have its value
    v_master = await cache.pop_from_queue("q", namespace="test")
    assert v_master == "master_a"
    assert await cache.pop_from_queue("q", namespace="test") is None


@pytest.mark.asyncio
async def test_acquire_and_release_lock(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    acquired = await cache.attempt_distributed_lock("lock1")
    assert acquired is True
    # In memory mode, always returns True
    await cache.release_distributed_lock("lock1")


@pytest.mark.asyncio
async def test_lru_cache_simple(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    call_counter = {"count": 0}

    @cache.lru_cache(expire=2)
    async def add(a: int, b: int) -> int:
        call_counter["count"] += 1
        return a + b

    # First call, should compute and cache
    result1 = await add(1, 2)
    assert result1 == 3
    assert call_counter["count"] == 1

    # Second call, same args, should hit cache
    result2 = await add(1, 2)
    assert result2 == 3
    assert call_counter["count"] == 1  # No increment, cache hit

    # Different args, should compute again
    result3 = await add(2, 3)
    assert result3 == 5
    assert call_counter["count"] == 2


@pytest.mark.asyncio
async def test_lru_cache_with_kwargs(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    call_counter = {"count": 0}

    @cache.lru_cache(expire=2)
    async def concat(a: int, b: int = 0) -> str:
        call_counter["count"] += 1
        return f"{a}-{b}"

    r1 = await concat(1, b=2)
    r2 = await concat(1, b=2)
    r3 = await concat(1, b=3)
    assert r1 == "1-2"
    assert r2 == "1-2"
    assert r3 == "1-3"
    assert call_counter["count"] == 2  # Only two unique calls


@pytest.mark.asyncio
async def test_lru_cache_json_serialization_error(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    @cache.lru_cache(expire=2)
    async def unserializable(arg: int) -> object:
        class NotSerializable:
            pass

        return NotSerializable()  # not JSON serializable

    with pytest.raises(ValueError):
        await unserializable(1)


@pytest.mark.asyncio
async def test_lru_cache_tenant_isolation(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    call_counter = {"count": 0}

    @cache.lru_cache(expire=2)
    async def add(a: int, b: int) -> int:
        call_counter["count"] += 1
        return a + b

    # Set in master tenant (default)
    result_master = await add(1, 2)
    assert result_master == 3
    assert call_counter["count"] == 1

    # Should hit cache in master tenant
    result_master2 = await add(1, 2)
    assert result_master2 == 3
    assert call_counter["count"] == 1

    # Switch to another tenant
    from brickworks.core.auth.authcontext import AuthContext

    async with AuthContext(tenant_schema="test_schema"):
        # Should NOT hit cache, should call function again
        result_tenant = await add(1, 2)
        assert result_tenant == 3
        assert call_counter["count"] == 2
        # Should hit cache in tenant
        result_tenant2 = await add(1, 2)
        assert result_tenant2 == 3
        assert call_counter["count"] == 2

    # Back to master tenant, should still hit master cache
    result_master3 = await add(1, 2)
    assert result_master3 == 3
    assert call_counter["count"] == 2


@pytest.mark.asyncio
async def test_set_add_and_members(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    # Add to set
    added = await cache.add_to_set("myset", "a", "b", "c")
    assert added == 3
    # Add duplicate and new
    added2 = await cache.add_to_set("myset", "b", "d")
    assert added2 == 1  # Only 'd' is new
    # Get members
    members = await cache.get_set_members("myset")
    assert members == {"a", "b", "c", "d"}
    # clean up
    await cache.remove_from_set("myset", "a", "b", "c", "d")


@pytest.mark.asyncio
async def test_set_remove_and_membership(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    await cache.add_to_set("myset2", "x", "y", "z")
    # Remove one
    removed = await cache.remove_from_set("myset2", "y")
    assert removed == 1
    # Remove non-existent
    removed2 = await cache.remove_from_set("myset2", "not_in_set")
    assert removed2 == 0
    # Membership
    assert await cache.is_set_member("myset2", "x") is True
    assert await cache.is_set_member("myset2", "y") is False
    # Remove all
    await cache.remove_from_set("myset2", "x", "z")
    members = await cache.get_set_members("myset2")
    assert members == set()


@pytest.mark.asyncio
async def test_set_tenant_isolation(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    await cache.add_to_set("tenants", "foo", namespace="test", master_tenant=True)
    async with AuthContext(tenant_schema="test_schema"):
        # Should not see master tenant's set
        assert await cache.get_set_members("tenants", namespace="test") == set()
        await cache.add_to_set("tenants", "bar", namespace="test")
        assert await cache.is_set_member("tenants", "bar", namespace="test") is True
        assert await cache.is_set_member("tenants", "foo", namespace="test") is False
        # Master tenant's set should be unchanged
        assert await cache.get_set_members("tenants", namespace="test", master_tenant=True) == {"foo"}
    # Back to master tenant
    assert await cache.get_set_members("tenants", namespace="test") == {"foo"}

    # clean up
    await cache.remove_from_set("tenants", "foo", namespace="test")
    async with AuthContext(tenant_schema="test_schema"):
        await cache.remove_from_set("tenants", "bar", namespace="test")


@pytest.mark.asyncio
async def test_index_add_and_cleanup(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    # Add two keys with the same index
    await cache.set_key("k1", "v1", namespace="test", indices=["myindex"])
    await cache.set_key("k2", "v2", namespace="test", indices=["myindex"])
    # Both keys should be listed by index
    keys = await cache.list_keys_by_index("myindex", namespace="test")
    assert set(keys) == {"k1", "k2"}

    # Delete one key
    await cache.delete_key("k1", namespace="test")
    # The index should be cleaned up, only k2 remains
    keys_after = await cache.list_keys_by_index("myindex", namespace="test")
    assert set(keys_after) == {"k2"}

    # Delete the other key
    await cache.delete_key("k2", namespace="test")
    # Now the index should be empty
    keys_final = await cache.list_keys_by_index("myindex", namespace="test")
    assert keys_final == []


@pytest.mark.asyncio
async def test_index_tenant_isolation(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    # Add key in master tenant
    await cache.set_key("k1", "v1", namespace="test", indices=["myindex"])
    # Switch to tenant and add another key with the same index
    async with AuthContext(tenant_schema="test_schema"):
        await cache.set_key("k2", "v2", namespace="test", indices=["myindex"])
        # Only k2 should be visible in tenant
        keys_tenant = await cache.list_keys_by_index("myindex", namespace="test")
        assert set(keys_tenant) == {"k2"}
        # Master tenant's key should not be visible
        assert "k1" not in keys_tenant
    # Back to master tenant, only k1 should be visible
    keys_master = await cache.list_keys_by_index("myindex", namespace="test")
    assert set(keys_master) == {"k1"}
    # Clean up
    await cache.delete_key("k1", namespace="test")
    async with AuthContext(tenant_schema="test_schema"):
        await cache.delete_key("k2", namespace="test")


@pytest.mark.asyncio
async def test_lru_cache_clear(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    call_counter = {"count": 0}

    @cache.lru_cache(expire=10)
    async def add(a: int, b: int) -> int:
        call_counter["count"] += 1
        return a + b

    # Call and cache a value
    result1 = await add(1, 2)
    assert result1 == 3
    assert call_counter["count"] == 1

    # Should hit cache
    result2 = await add(1, 2)
    assert result2 == 3
    assert call_counter["count"] == 1

    # Clear the cache
    await add.cache_clear()

    # Should call function again (cache miss)
    result3 = await add(1, 2)
    assert result3 == 3
    assert call_counter["count"] == 2


@pytest.mark.asyncio
async def test_lru_cache_clear_tenant_isolation(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    call_counter = {"count": 0}

    @cache.lru_cache(expire=2)
    async def add(a: int, b: int) -> int:
        call_counter["count"] += 1
        return a + b

    # Call and cache in master tenant
    result_master = await add(1, 2)
    assert result_master == 3
    assert call_counter["count"] == 1

    # Should hit cache in master tenant
    result_master2 = await add(1, 2)
    assert result_master2 == 3
    assert call_counter["count"] == 1

    # Switch to tenant and call (should miss cache)
    async with AuthContext(tenant_schema="test_schema"):
        result_tenant = await add(1, 2)
        assert result_tenant == 3
        assert call_counter["count"] == 2
        # Should hit cache in tenant
        result_tenant2 = await add(1, 2)
        assert result_tenant2 == 3
        assert call_counter["count"] == 2
        # Clear cache in tenant
        await add.cache_clear()
        # Should call function again (cache miss in tenant)
        result_tenant3 = await add(1, 2)
        assert result_tenant3 == 3
        assert call_counter["count"] == 3
        # Should hit cache in tenant again
        result_tenant4 = await add(1, 2)
        assert result_tenant4 == 3
        assert call_counter["count"] == 3

    # Back to master tenant, should still hit master cache (not affected by tenant clear)
    result_master3 = await add(1, 2)
    assert result_master3 == 3
    assert call_counter["count"] == 3

    # Now clear cache in master tenant
    await add.cache_clear()
    result_master4 = await add(1, 2)
    assert result_master4 == 3
    assert call_counter["count"] == 4


@pytest.mark.asyncio
async def test_lock_exclusive(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    acquired1 = await cache.attempt_distributed_lock("locktest", ttl=2)
    assert acquired1 is True
    acquired2 = await cache.attempt_distributed_lock("locktest", ttl=2)
    assert acquired2 is False
    await cache.release_distributed_lock("locktest")
    acquired3 = await cache.attempt_distributed_lock("locktest", ttl=2)
    assert acquired3 is True
    await cache.release_distributed_lock("locktest")


@pytest.mark.asyncio
async def test_lock_expiry(app: TestApp, cache_backend: TCacheBackendFixture) -> None:
    import asyncio

    acquired1 = await cache.attempt_distributed_lock("lockexpire", ttl=1)
    assert acquired1 is True
    acquired2 = await cache.attempt_distributed_lock("lockexpire", ttl=1)
    assert acquired2 is False

    await asyncio.sleep(1.1)
    acquired3 = await cache.attempt_distributed_lock("lockexpire", ttl=1)
    assert acquired3 is True
    await cache.release_distributed_lock("lockexpire")



# pylint: disable=no-member
import asyncio
import logging
from logging.config import fileConfig

from alembic import context
from sqlalchemy import Connection, pool, schema, select, text
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from brickworks.core import db
from brickworks.core.db import get_db_url, reset_sessionmaker
from brickworks.core.models.base_dbmodel import Base
from brickworks.core.settings import settings

config = context.config


config.set_main_option("sqlalchemy.url", get_db_url())

logger = logging.getLogger("env_py")

# Interpret the config file for Python logging.
# This line sets up loggers basically.
# only set up logging if it isn't already configured
if config.config_file_name is not None and not logging.getLogger().handlers:
    fileConfig(config.config_file_name)

target_metadata = Base.metadata

target_metadata.naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


async def get_schema_list() -> list[str]:
    """
    Fetch a list of all tenant schemas from the database.
    This requires that the master schema is already created, or this will crash!
    """
    from brickworks.core.models.tenant_model import TenantModel

    async with db(schema=settings.MASTER_DB_SCHEMA):
        query = select(TenantModel.schema)
        result = await db.session.execute(query)
        return [row[0] for row in result]


def do_run_migrations(connection: Connection | None) -> None:
    context.configure(connection=connection, target_metadata=target_metadata, render_as_batch=True)

    with context.begin_transaction():
        context.run_migrations()


async def create_schema(schema_name: str, conntactable: AsyncEngine) -> None:
    try:
        async with conntactable.connect() as connection:
            await connection.execute(schema.CreateSchema(schema_name))
            await connection.commit()
        logger.info(f"Created schema {schema_name}")
    except ProgrammingError:
        pass  # schema already exists, ignore the error


async def run_migrations_online() -> None:
    """Run migrations in 'online' mode.
    In this scenario we need to create an Engine
    and associate a connection with the context.
    """
    reset_sessionmaker()  # reset sessionmaker to ensure the engine is running in the correct event loop

    # check which operation we are performing
    # we pass them as x-arguments
    try:
        is_migration = context.get_x_argument(as_dictionary=True).get("mode") == "migrate"
        schema = context.get_x_argument(as_dictionary=True).get("schema", "")
    except AttributeError:
        # if no x-arguments are provided an AttributeError is raised
        # so we assume the defaults
        is_migration = False
        schema = ""
    # get the database connection...
    url: str = config.get_main_option("sqlalchemy.url") or ""
    if not url or not url.startswith("postgresql+asyncpg://"):
        raise ValueError("Invalid database URL. It must start with 'postgresql+asyncpg://'.")
    connectable = create_async_engine(url, poolclass=pool.NullPool)

    if is_migration:
        if schema:
            # if a schema is provided, we run migrations for that schema only
            # this would be the case after a new tenant is created
            await run_migrations_with_schema(connectable, schema)
            return
        # if no schema is provided, we run migrations for all schemas

        # run migration for master first, because we need to exist, otherwise
        # attempting to fetch the list of schemas will crash
        await run_migrations_with_schema(connectable, settings.MASTER_DB_SCHEMA)

        # now we can fetch the list of schemas
        schema_names = await get_schema_list()
        if settings.MASTER_DB_SCHEMA in schema_names:
            # remove the master schema from the list of schemas to migrate
            # because we already ran migrations for it
            schema_names.remove(settings.MASTER_DB_SCHEMA)

        # migrate all other schemas
        for schema_name in schema_names:
            await run_migrations_with_schema(connectable, schema_name)
    else:
        # might be creating a new revision or checking for changes
        # so we only run migrations for the master schema
        await run_migrations_with_schema(connectable, settings.MASTER_DB_SCHEMA)

    await connectable.dispose()


async def run_migrations_with_schema(connectable: AsyncEngine, schema_name: str) -> None:
    await create_schema(schema_name, connectable)
    logger.info(f"Running migrations for schema {schema_name}")
    async with connectable.connect() as connection:
        await connection.execution_options(schema_translate_map={None: schema_name})
        await connection.execute(text(f'set search_path to "{schema_name}"'))
        await connection.commit()
        # make use of non-supported SQLAlchemy attribute to ensure
        # the dialect reflects tables in terms of the current tenant name
        connection.dialect.default_schema_name = schema_name
        await connection.run_sync(do_run_migrations)


asyncio.run(run_migrations_online())



from brickworks.core.models.role_model import RoleModel
from brickworks.core.models.user_model import UserModel
from tests.conftest import TestApp


async def test_role_add_remove(app: TestApp) -> None:
    role1 = await RoleModel(role_name="role1").persist()
    user_alice = await UserModel(
        sub="alice", given_name="Alice", family_name="Smith", name="Alice Smith", email="alice@example.com"
    ).persist()

    # add role by name
    await user_alice.add_role("role1")

    # adding the same role again should not cause issues
    await user_alice.add_role("role1")

    assert await user_alice.has_role("role1")

    # remove role
    await user_alice.remove_role("role1")

    assert not await user_alice.has_role("role1")

    # add role by object
    await user_alice.add_role(role1)
    assert await user_alice.has_role(role1)

    # remove role by object
    await user_alice.remove_role(role1)
    assert not await user_alice.has_role(role1)


async def test_give_role_permission(app: TestApp) -> None:
    await RoleModel(role_name="role1").persist()
    user_bob = await UserModel(
        sub="bob", given_name="Bob", family_name="Smith", name="Bob Smith", email="bob@example.com"
    ).persist()

    await user_bob.give_role_permission("role1", "read")
    # adding the same permission again should not cause issues
    await user_bob.give_role_permission("role1", "read")

    assert await user_bob.has_role_permission("role1", "read")

    # remove role permission
    await user_bob.remove_role_permission("role1", "read")
    # removing the same permission again should not cause issues
    await user_bob.remove_role_permission("role1", "read")
    assert not await user_bob.has_role_permission("role1", "read")



from sqlalchemy import func, select
from sqlalchemy.orm import aliased

from brickworks.core.acl.policies import RoleBasedAccessPolicy
from brickworks.core.auth.authcontext import AuthContext
from brickworks.core.models.base_view import BaseView
from brickworks.core.models.role_model import RoleModel, user_role_table
from brickworks.core.models.user_model import UserModel
from tests.conftest import TestApp
from tests.core.utils import create_test_user

user_alias = aliased(UserModel)
role_alias = aliased(RoleModel)


class UserRoleViewTest(BaseView):
    """
    View displaying user roles.
    """

    role_name: str
    user_name: str

    __select__ = (
        select(
            role_alias.role_name.label("role_name"),
            user_alias.name.label("user_name"),
        )
        .join(role_alias.users)
        .join(user_alias)
    )
    __policy_model_class__ = UserModel
    __policies__ = [RoleBasedAccessPolicy("admin")]


class RolesPerUserViewTest(BaseView):
    """
    View displaying each user and counts the number of roles they have.
    """

    user_name: str
    role_count: int

    __select__ = (
        select(
            user_alias.name.label("user_name"),
            func.count(user_role_table.c.role_uuid).label("role_count"),
        )
        .select_from(user_alias)
        .join(user_role_table, user_alias.uuid == user_role_table.c.user_uuid)
        .group_by(user_alias.name)
    )
    __policy_model_class__ = UserModel
    __policies__ = [RoleBasedAccessPolicy("admin")]


async def test_user_role_view(app: TestApp) -> None:
    """
    Test the UserRoleView class.
    """
    # Create some test users
    alice = await create_test_user("Alice", "Smith")
    bob = await create_test_user("Bob", "Johnson")
    charlie = await create_test_user("Charlie", "Brown")

    # Create some test roles
    role_admin = await RoleModel(role_name="admin").persist()
    role_user = await RoleModel(role_name="user").persist()

    # Assign roles to users
    await alice.add_role(role_admin)
    await alice.add_role(role_user)
    await bob.add_role(role_user)
    await charlie.add_role(role_user)

    # Query the user role view
    user_roles = await UserRoleViewTest.get_list()
    assert len(user_roles) >= 4
    alice_roles = [role for role in user_roles if role.user_name == "Alice Smith"]
    assert len(alice_roles) == 2

    # Query the user count view
    user_count = await RolesPerUserViewTest.get_list()
    assert len(user_count) >= 3
    alice_count = next((user for user in user_count if user.user_name == "Alice Smith"), None)
    assert alice_count is not None
    assert alice_count.role_count == 2


async def test_user_role_view_with_policies(app: TestApp) -> None:
    """
    Test the UserRoleView class with policies.
    """
    # Create some test users
    alice = await create_test_user("Alice", "Smith")
    bob = await create_test_user("Bob", "Johnson")
    charlie = await create_test_user("Charlie", "Brown")
    # Create some test roles
    role_admin = await RoleModel(role_name="admin").persist()
    role_user = await RoleModel(role_name="user").persist()
    # Assign roles to users
    await alice.add_role(role_user)
    await bob.add_role(role_user)
    await charlie.add_role(role_user)

    async with AuthContext(alice.uuid):
        # Alice does not have the admin role, so she should not see any results
        user_roles = await UserRoleViewTest.get_list_with_policies()
        assert len(user_roles) == 0
        user_count = await RolesPerUserViewTest.get_list_with_policies()
        assert len(user_count) == 0

        await alice.add_role(role_admin)

        # Now Alice has the admin role, so she should see all users
        user_roles = await UserRoleViewTest.get_list_with_policies()
        assert len(user_roles) >= 4

        user_count = await RolesPerUserViewTest.get_list_with_policies()
        assert len(user_count) >= 3
        alice_count = next((user for user in user_count if user.user_name == "Alice Smith"), None)
        assert alice_count is not None
        assert alice_count.role_count == 2



import importlib
from typing import Any


def import_object_from_path(attribute_path: str) -> Any:  # noqa: ANN401
    """
    import an object from a path like "module.submodule.object"
    """
    split_path = attribute_path.split(".")
    attribute_name = split_path.pop()
    module_name = ".".join(split_path)

    module = importlib.import_module(module_name)
    obj = getattr(module, attribute_name)

    return obj



from brickworks.core.models.user_model import UserModel


async def create_test_user(user_given_name: str, user_family_name: str) -> UserModel:
    """
    Create a test user with the given name and family name.
    """
    test_user = await UserModel(
        sub=user_given_name.lower(),
        name=f"{user_given_name} {user_family_name}",
        given_name=user_given_name,
        family_name=user_family_name,
        email=f"{user_given_name}@example.com",
    ).persist()
    return test_user



import asyncio
import functools
import hashlib
import json
import logging
import time
from collections.abc import Awaitable, Callable, Set
from dataclasses import dataclass
from typing import ParamSpec, Protocol, TypeVar, cast, runtime_checkable

from redis.asyncio import Redis

from brickworks.core.settings import settings

logger = logging.getLogger(__name__)


P = ParamSpec("P")  # parameters of cached function
R = TypeVar("R")  # return type of cached function
R_co = TypeVar("R_co", covariant=True)

NAMESPACE_LRU_CACHE = "LRU_CACHE"


@dataclass
class _CacheEntry:
    expiration_time: int
    value: str


@runtime_checkable
class LruCacheWrapper(Protocol[P, R_co]):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Awaitable[R_co]: ...
    async def cache_clear(self) -> None: ...


class BrickworksCache:
    """Caches keys either in memory or in redis"""

    _redis_client: Redis
    _memory_cache: dict[str, _CacheEntry] = {}
    _memory_queue: dict[str, list[str]] = {}
    _memory_sets: dict[str, set[str]] = {}
    _memory_cache_lock: asyncio.Lock = asyncio.Lock()
    PREFIX = "brickworks"

    def __init__(self) -> None:
        if settings.USE_REDIS:
            self._redis_client = Redis(
                host=settings.REDIS_HOST,
                port=settings.REDIS_PORT,
                db=settings.REDIS_DB,
                password=settings.REDIS_PASSWORD,
            )
        else:
            logger.warning("Using memory cache. This is not recommended for production use.")

    def _generate_key(self, key: str, namespace: str, master_tenant: bool) -> str:
        tenant: str = settings.MASTER_DB_SCHEMA
        if not master_tenant:
            from brickworks.core.auth.authcontext import auth_context

            tenant_schema = auth_context.tenant_schema
            if not tenant_schema:
                raise ValueError("Tried to access current tenant before auth context has a tenant set")
            tenant = tenant_schema

        # Validate that PREFIX, tenant, and namespace do not contain ':'
        if ":" in self.PREFIX:
            raise ValueError("PREFIX should not contain ':'")
        if ":" in tenant:
            raise ValueError("Tenant should not contain ':'")
        if ":" in namespace:
            raise ValueError("Namespace should not contain ':'")
        if ":" in key:
            raise ValueError("Key should not contain ':'")
        return f"{self.PREFIX}:{tenant}:{namespace}:{key}"

    def _parse_key(self, key: str) -> tuple[str, str, str]:
        # separate the key into its components to retrieve tenant namespace and key
        split = key.split(":")
        if len(split) != 4:
            raise ValueError(f"key {key} does not follow the schema prefix:tenant:namespace:key")
        return (split[1], split[2], split[3])

    async def push_to_queue(
        self, queue: str, value: str, namespace: str = "default", master_tenant: bool = False
    ) -> None:
        """Pushes a value to a queue. If no namespace is given, the default namespace is used."""
        cache_key = self._generate_key(queue, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            # need to use cast here to help out pylance
            await cast(
                Awaitable[int],
                self._redis_client.rpush(cache_key, value),
            )
        else:
            if cache_key not in self._memory_queue:
                self._memory_queue[cache_key] = []
            self._memory_queue[cache_key].append(value)

    async def pop_from_queue(self, queue: str, namespace: str = "default", master_tenant: bool = False) -> str | None:
        """Pops a value from a queue. If no namespace is given, the default namespace is used."""
        cache_key = self._generate_key(queue, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            # need to use cast here to help out pylance
            res: bytes = await cast(Awaitable[bytes], self._redis_client.lpop(cache_key))
            return res.decode("utf-8") if res else None
        if cache_key not in self._memory_queue:
            return None
        if not self._memory_queue[cache_key]:
            return None
        return self._memory_queue[cache_key].pop(0)

    async def get_key(self, key: str, namespace: str = "default", master_tenant: bool = False) -> str | None:
        """Retrieves a key from the cache. If no namespace is given, the default namespace is used."""
        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            res = await self._redis_client.get(cache_key)
            return res.decode("utf-8") if res else None

        cache_entry = self._memory_cache.get(cache_key)
        if cache_entry is None:
            return None
        if cache_entry.expiration_time < int(time.time()):
            del self._memory_cache[cache_key]
            return None
        return cache_entry.value

    async def set_key(
        self,
        key: str,
        value: str,
        namespace: str = "default",
        expire: int = 3600 * 24 * 7,
        master_tenant: bool = False,
        indices: list[str] | None = None,
        nx: bool = False,
    ) -> bool:
        """Sets a key in the cache. If no namespace is given, the default namespace is used.
        Cache expiration time is 7 days by default.
        If nx is True, only set if key does not exist (like Redis SETNX).
        Returns True if the key was set, False otherwise.
        """
        if not isinstance(value, str):
            raise ValueError(f"Cached value must be a string. Got {type(value)}")

        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            result = await self._redis_client.set(cache_key, value, ex=expire, nx=nx)
            if not result:
                return False
        else:
            now = int(time.time())
            async with self._memory_cache_lock:
                entry = self._memory_cache.get(cache_key)
                if nx and entry is not None and entry.expiration_time > now:
                    return False
                self._memory_cache[cache_key] = _CacheEntry(now + expire, value)
        for index in indices or []:
            await self.add_to_index(index, key, namespace, master_tenant=master_tenant)
        return True

    async def delete_key(self, key: str, namespace: str = "default", master_tenant: bool = False) -> None:
        """Deletes a key from the cache. If no namespace is given, the default namespace is used."""
        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if await self.get_key(key, namespace, master_tenant=master_tenant) is None:
            return
        if settings.USE_REDIS:
            await self._redis_client.delete(cache_key)
        else:
            del self._memory_cache[cache_key]

    async def refresh_key(
        self,
        key: str,
        namespace: str = "default",
        expire: int = 3600 * 24 * 7,
        master_tenant: bool = False,
    ) -> None:
        """Refreshes the expiration time of a key in the cache."""
        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            await self._redis_client.expireat(cache_key, int(time.time()) + expire)
        else:
            cache_entry = self._memory_cache.get(cache_key)
            if cache_entry is None:
                return
            cache_entry.expiration_time = int(time.time()) + expire

    def delete_expired_entries(self) -> None:
        """Deletes all expired entries from the memory cache.
        Doesn't do anyting if redis is used.
        """
        if settings.USE_REDIS:
            return
        for key, cache_entry in self._memory_cache.items():
            if cache_entry.expiration_time < int(time.time()):
                del self._memory_cache[key]

    async def attempt_distributed_lock(self, lock_name: str, ttl: int = 10, master_tenant: bool = False) -> bool:
        """Attempts to acquire a distributed lock using Redis or in-memory cache.
        Returns True if the lock was acquired, False otherwise.
        ttl: The time-to-live (expiration) for the lock in seconds.
        """
        return await self.set_key(
            lock_name, "1", namespace="__locks__", expire=ttl, nx=True, master_tenant=master_tenant
        )

    async def release_distributed_lock(self, lock_name: str, master_tenant: bool = False) -> None:
        """Releases a distributed lock."""
        await self.delete_key(lock_name, namespace="__locks__", master_tenant=master_tenant)

    async def add_to_set(
        self,
        key: str,
        *values: str,
        namespace: str = "default",
        master_tenant: bool = False,
    ) -> int:
        """Add one or more members to a set. Returns the number of elements added."""
        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            return await cast(Awaitable[int], self._redis_client.sadd(cache_key, *values))
        s = self._memory_sets.setdefault(cache_key, set())
        before = len(s)
        s.update(values)
        return len(s) - before

    async def remove_from_set(
        self,
        key: str,
        *values: str,
        namespace: str = "default",
        master_tenant: bool = False,
    ) -> int:
        """Remove one or more members from a set. Returns the number of elements removed."""
        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            return await cast(Awaitable[int], self._redis_client.srem(cache_key, *values))
        s = self._memory_sets.get(cache_key, set())
        before = len(s)
        for v in values:
            s.discard(v)
        removed = before - len(s)
        if not s:
            self._memory_sets.pop(cache_key, None)
        return removed

    async def get_set_members(
        self,
        key: str,
        namespace: str = "default",
        master_tenant: bool = False,
    ) -> Set[str]:
        """Get all members of a set. Returns a set of strings."""
        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            members = await cast(Awaitable[set[str]], self._redis_client.smembers(cache_key))
            return {m.decode("utf-8") if isinstance(m, bytes) else str(m) for m in members}
        return set(self._memory_sets.get(cache_key, set()))

    async def is_set_member(
        self,
        key: str,
        value: str,
        namespace: str = "default",
        master_tenant: bool = False,
    ) -> bool:
        """Check if a value is a member of a set."""
        cache_key = self._generate_key(key, namespace=namespace, master_tenant=master_tenant)
        if settings.USE_REDIS:
            result = await cast(Awaitable[int], self._redis_client.sismember(cache_key, value))
            return bool(result)
        return value in self._memory_sets.get(cache_key, set())

    async def add_to_index(
        self,
        index: str,
        key: str,
        namespace: str = "default",
        master_tenant: bool = False,
    ) -> None:
        """Adds a key to an index and registers the index in the central index set."""
        index_key = f"__index__.{index}"
        # Central set of all indexes (per namespace/tenant)
        central_index_set = "__indexes__"
        await self.add_to_set(
            central_index_set,
            self._generate_key(index_key, namespace=namespace, master_tenant=master_tenant),
            namespace="default",
            master_tenant=master_tenant,
        )
        # Add the key to the index set
        await self.add_to_set(
            index_key,
            self._generate_key(key, namespace=namespace, master_tenant=master_tenant),
            namespace=namespace,
            master_tenant=master_tenant,
        )

    async def _clean_index(self, cache_key: str) -> None:
        """Removes keys from the index that don't exist anymore"""
        if settings.USE_REDIS:
            # get all keys in the index
            keys = await cast(Awaitable[set[str]], self._redis_client.smembers(cache_key))
            for key in keys:
                # check if the key exists, if not delete from the index
                if not await self._redis_client.exists(key):
                    await cast(Awaitable[int], self._redis_client.srem(cache_key, key))
        else:
            # get all keys in the index
            keys = self._memory_sets.get(cache_key, set()).copy()
            for key in keys:
                if key not in self._memory_cache:
                    self._memory_sets[cache_key].discard(key)
            if not self._memory_sets[cache_key]:
                self._memory_sets.pop(cache_key, None)

    async def clean_all_indices(self) -> None:
        index_cache_keys = await self.get_set_members("__indexes__", namespace="default")
        for cache_key in index_cache_keys:
            await self._clean_index(cache_key)

    async def list_keys_by_index(
        self, index: str, namespace: str = "default", master_tenant: bool = False
    ) -> list[str]:
        index_key = f"__index__.{index}"
        await self._clean_index(self._generate_key(index_key, namespace, master_tenant))
        cache_keys = await self.get_set_members(index_key, namespace, master_tenant)
        return [self._parse_key(cache_key)[2] for cache_key in cache_keys]

    def lru_cache(
        self, expire: int = 3600 * 24 * 7, master_tenant: bool = False
    ) -> Callable[[Callable[P, Awaitable[R]]], LruCacheWrapper[P, R]]:
        """Decorator to cache async function results using BrickworksCache.
        Only works with JSON-serializable arguments and return values.
        """

        def decorator(func: Callable[P, Awaitable[R]]) -> LruCacheWrapper[P, R]:
            fqpn = f"{func.__module__}.{func.__qualname__}"

            @functools.wraps(func)
            async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                # Serialize args and kwargs to a stable string
                try:
                    key_data = json.dumps({"args": args, "kwargs": kwargs}, sort_keys=True)
                except Exception as e:
                    raise ValueError(f"Arguments to {fqpn} are not JSON serializable: {e}") from e
                key_hash = hashlib.sha256(key_data.encode()).hexdigest()
                cache_key = f"{fqpn}-{key_hash}"
                cached = await self.get_key(cache_key, namespace=NAMESPACE_LRU_CACHE, master_tenant=master_tenant)
                if cached is not None:
                    return cast(R, json.loads(cached))
                result = await func(*args, **kwargs)
                try:
                    result_json = json.dumps(result, sort_keys=True)
                except Exception as e:
                    raise ValueError(f"Return value of {fqpn} is not JSON serializable: {e}") from e
                await self.set_key(
                    cache_key,
                    result_json,
                    namespace=NAMESPACE_LRU_CACHE,
                    expire=expire,
                    master_tenant=master_tenant,
                    indices=[fqpn],
                )
                return result

            async def cache_clear(master_tenant: bool = False) -> None:
                keys = await self.list_keys_by_index(fqpn, namespace=NAMESPACE_LRU_CACHE, master_tenant=master_tenant)
                for key in keys:
                    await self.delete_key(key, namespace=NAMESPACE_LRU_CACHE, master_tenant=master_tenant)

            wrapper.__setattr__("cache_clear", cache_clear)
            return wrapper  # type: ignore[return-value]

        return decorator


cache = BrickworksCache()



import uuid

from sqlalchemy import Boolean, Integer, String
from sqlalchemy.orm import Mapped, mapped_column

from brickworks.core.exceptions import NotFoundException
from brickworks.core.models.base_dbmodel import UUID_LENGTH, BaseDBModel
from brickworks.core.module_loader import get_models_by_fqpn

PUBLIC_BLOB_ID_PREFIX = "public-"


class WithFilesMixin(BaseDBModel):
    __abstract__ = True

    async def on_file_created(self, file_record: "FileModel") -> None:
        """
        This method is called after a file is created. At this stage only the file record is
        created, but no file is uploaded yet.
        You can still raise an exception here, which will rollback the transaction and thus
        prevent the file from being created.
        """
        pass


class FileModel(BaseDBModel):
    __tablename__ = "core_files"
    blob_id: Mapped[str] = mapped_column(String(255), unique=True, nullable=False)
    file_name: Mapped[str] = mapped_column(String(255), nullable=False)
    parent_uuid: Mapped[str] = mapped_column(String(UUID_LENGTH), nullable=False)
    parent_fqpn: Mapped[str] = mapped_column(String(255), nullable=False)
    size: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    mimetype: Mapped[str | None] = mapped_column(String(255), nullable=True, default=None)
    finalized: Mapped[bool] = mapped_column(Boolean, nullable=False, default=False)

    @classmethod
    async def create(
        cls,
        user_uuid: str,
        file_name: str,
        parent_uuid: str,
        parent_fqpn: str,
        size: int = 0,
        mimetype: str | None = None,
        public: bool = False,
    ) -> "FileModel":
        """
        Create a new file record in the database.
        Files need to be attached to a parent object, which is identified by the parent_uuid and
        the fully qualified path name of the parent object.

        By default, blobs are not public, meaning a presigned URL is required to access them.
        Public blobs can be accessed directly through a static URL.
        This is useful for files that are meant to be shared publicly, such as images on a website.
        """
        parent_object_class = get_models_by_fqpn().get(parent_fqpn)
        if not parent_object_class:
            raise ValueError(f"Parent object class not found for fqpn: {parent_fqpn}")
        # check that the parent class is a subclass of WithFilesMixin
        if not issubclass(parent_object_class, WithFilesMixin):
            raise ValueError(f"Parent object class {parent_fqpn} is not a subclass of WithFilesMixin")

        # check that the parent object exists
        parent_object = await parent_object_class.get_one_or_none(uuid=parent_uuid)
        if not parent_object:
            raise NotFoundException(f"Parent object with UUID {parent_uuid} of class {parent_fqpn} not found")

        file_record = await cls(
            blob_id=cls.generate_blob_id(public=public),
            file_name=file_name,
            parent_uuid=parent_uuid,
            parent_fqpn=parent_fqpn,
            size=size,
            mimetype=mimetype,
        ).persist()
        await parent_object.on_file_created(file_record)
        return file_record

    @staticmethod
    def generate_blob_id(public: bool = False) -> str:
        """
        Generate a unique blob ID for the file.
        This is a placeholder implementation. In a real application, you would use a more robust
        method to generate unique IDs, such as UUIDs or a combination of timestamp and random
        string.
        """

        blob_id = str(uuid.uuid4())
        if public:
            blob_id = PUBLIC_BLOB_ID_PREFIX + blob_id
        return blob_id



import logging

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from brickworks.core.auth.authcontext import AuthContextMiddleware
from brickworks.core.auth.csrf import CSRFMiddleware
from brickworks.core.auth.session import SessionMiddleware
from brickworks.core.db import DBSessionMiddleware
from brickworks.core.module_loader import load_modules
from brickworks.core.utils.importer import import_object_from_path

logging.basicConfig(level=logging.INFO)


def create_app(for_testing: bool = False) -> FastAPI:
    app_base = FastAPI()
    app_api = FastAPI()

    # Middleware is executed in reverse order, so we first add the middleware that should be executed last
    # Middleware loaded from bricks should be executed last, so we add it first
    # (to ensure it has access to sessions and the database)
    for brick in load_modules():
        _add_routers(app_api, brick.routers)
        _add_middlewares(app_api, brick.middlewares)

    if not for_testing:
        # if we run the app with testclient we will create database sessions ourselves, so we can roll back
        app_api.add_middleware(DBSessionMiddleware)

    app_api.add_middleware(AuthContextMiddleware)  # needs to be before session middleware

    app_api.add_middleware(CSRFMiddleware)  # needs to be before session middleware
    app_api.add_middleware(
        SessionMiddleware,
        cookie_name="session",
        max_age=14 * 24 * 60 * 60,
        same_site="lax",
        https_only=False,
        domain=None,
    )
    app_api.add_middleware(
        CORSMiddleware,
        allow_origins=["http://localhost:3000", "http://localhost:8000"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    # app_base.add_event_handler("startup", start_services)

    app_base.mount("/backend/api", app_api)
    return app_base


def _add_routers(app: FastAPI, routers: list[str]) -> None:
    for router_path in routers:
        router = import_object_from_path(router_path)
        app.include_router(router)


def _add_middlewares(app: FastAPI, middlewares: list[str]) -> None:
    for middleware_path in middlewares:
        middleware = import_object_from_path(middleware_path)
        app.add_middleware(middleware)


if __name__ == "__main__":
    app = create_app()
    uvicorn.run(app, host="0.0.0.0", port=8000)  # nosec



from typing import Any

from fastapi import HTTPException


class CustomException(HTTPException):
    headers: dict[str, str] | None = None
    msg: str
    type: str
    loc: list[str] | None = []

    def __init__(self, msg: str | None = None) -> None:
        self.msg = msg or self.msg  # can override the default message
        self.detail: Any = [
            {
                "loc": self.loc,
                "msg": self.msg,
                "type": self.type,
            }
        ]
        super().__init__(status_code=self.status_code, detail=self.detail, headers=self.headers)


class NotFoundException(CustomException):
    status_code: int = 404


class UnauthorizedException(CustomException):
    status_code: int = 401


class ForbiddenException(CustomException):
    status_code: int = 403


class DuplicateException(CustomException):
    status_code: int = 409
    type: str = "duplicate_error"
    msg: str = "Duplicate entry found"



from typing import Any, cast

from sqlalchemy.sql._typing import (
    _ColumnExpressionArgument,
    _ColumnExpressionOrStrLabelArgument,
)

TypeOrderBy = _ColumnExpressionOrStrLabelArgument[Any] | list[_ColumnExpressionOrStrLabelArgument[Any]]
TypeWhereClause = _ColumnExpressionArgument[bool]

AlwaysTrueWhereClause = cast(TypeWhereClause, 1 == 1)
AlwaysFalseWhereClause = cast(TypeWhereClause, 1 == 2)  # type: ignore[comparison-overlap]



import os
from collections.abc import Awaitable, Callable
from contextvars import ContextVar, Token
from dataclasses import dataclass, field
from types import TracebackType
from typing import Any

from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from brickworks.core.settings import settings

_SESSION: async_sessionmaker[AsyncSession] | None = None


@dataclass
class DBSessionContext:
    session: AsyncSession
    on_post_committed: list[Callable[[], Awaitable[Any]]] = field(
        default_factory=list
    )  # list of async functions called after committing the session


_session_context_var: ContextVar[DBSessionContext | None] = ContextVar("_session_context", default=None)

CURRENT_FILE_PATH = os.path.abspath(__file__)


def get_db_url() -> str:
    return f"postgresql+asyncpg://{settings.DB_USER}:{settings.DB_PASSWORD}@{settings.DB_HOST}/{settings.DB_NAME}"


class MissingSessionError(Exception):
    """Exception raised for when the user tries to access a database session before it is created."""

    def __init__(self) -> None:
        msg = """
        No session found! Either you are not currently in a request context,
        or you need to manually create a session context by using a `db` instance as
        a context manager e.g.:
        async with db():
            await db.session.execute(foo.select()).fetchall()
        """

        super().__init__(msg)


class SessionNotInitialisedError(Exception):
    """Exception raised when the user creates a new DB session without first initialising it."""

    def __init__(self) -> None:
        msg = """
        Session not initialised! Ensure that DBSessionMiddleware has been initialised before
        attempting database access.
        """

        super().__init__(msg)


def set_sessionmaker() -> None:
    engine_args = {
        "pool_size": 10,
        "max_overflow": 20,
        "pool_pre_ping": True,
    }

    session_args: dict[str, Any] = {}
    engine = create_async_engine(get_db_url(), **engine_args)

    global _SESSION  # pylint: disable=global-statement
    _SESSION = async_sessionmaker(engine, class_=AsyncSession, expire_on_commit=False, **session_args)


def reset_sessionmaker() -> None:
    """
    Needed if you start a new event loop and want to ensure that the sessionmaker is re-initialised.
    Otherwise you might see errors like "Future attached to a different loop."
    """
    global _SESSION  # pylint: disable=global-statement
    _SESSION = None
    _session_context_var.set(None)  # reset the context variable as well


class DBSessionMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)
        set_sessionmaker()

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        # TODO: get schema from request
        async with db(commit_on_exit=True):
            response = await call_next(request)
            if response.status_code >= 400:
                await db.session.rollback()
            return response


class DBSessionMeta(type):
    # using this metaclass means that we can access db.session as a property at a class level,
    # rather than db().session
    @property
    def session(cls) -> AsyncSession:
        """Return an instance of Session local to the current async context."""
        if _SESSION is None:
            raise SessionNotInitialisedError

        return cls.session_context.session

    def add_post_commit_callback(cls, callback: Callable[[], Awaitable[Any]]) -> None:
        """
        Add an async function that should be called after the session has been committed.

        Post-commit callbacks are useful for actions that should only happen if the transaction is successful.
        For example, if you have a cache that depends on database objects, updating the cache before commit
        could lead to inconsistencies if the transaction is rolled back. By registering a post-commit callback,
        you ensure that cache updates (or other side effects) only occur after the database changes are committed
        and guaranteed to persist.
        """
        cls.session_context.on_post_committed.append(callback)

    @property
    def session_context(cls) -> DBSessionContext:
        """Return the current session context."""
        context = _session_context_var.get()
        if context is None:
            raise MissingSessionError
        return context


class DBSession(metaclass=DBSessionMeta):
    def __init__(
        self,
        schema: str = settings.MASTER_DB_SCHEMA,
        session_args: dict[str, Any] | None = None,
        commit_on_exit: bool = False,
    ) -> None:
        self.token: Token[DBSessionContext | None] | None = None
        self.schema = schema
        self.session_args = session_args or {}
        self.commit_on_exit = commit_on_exit

    async def _init_session(self) -> None:
        assert _SESSION is not None  # nosec  we just do the assert to help out VSCode
        session: AsyncSession = _SESSION(**self.session_args)
        execution_options = {"schema_translate_map": {None: self.schema}}

        await session.connection(execution_options=execution_options)
        context = DBSessionContext(session=session)
        self.token = _session_context_var.set(context)

    async def __aenter__(self) -> type["DBSession"]:
        if not isinstance(_SESSION, async_sessionmaker):
            set_sessionmaker()

        await self._init_session()
        return type(self)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        context = _session_context_var.get()
        if not context:
            raise MissingSessionError
        session = context.session
        if exc_type is not None:
            await session.rollback()

        if self.commit_on_exit:
            await session.commit()
            for callback in context.on_post_committed:
                await callback()

        await session.close()
        _session_context_var.reset(self.token)  # type: ignore


db: DBSessionMeta = DBSession



import logging
from collections.abc import Sequence
from datetime import datetime
from typing import Any, ClassVar, Literal, Self, TypeVar, final
from uuid import uuid4

from sqlalchemy import DateTime, Select, String, and_, delete, insert, or_, select
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column
from sqlalchemy.sql._typing import _ColumnExpressionOrStrLabelArgument

from brickworks.core.acl.base_policy import BasePolicy
from brickworks.core.auth.authcontext import auth_context
from brickworks.core.db import db
from brickworks.core.signals import BaseSignal, signals
from brickworks.core.utils.sqlalchemy import TypeWhereClause
from brickworks.core.utils.timeutils import now_utc

logger = logging.getLogger(__name__)
UUID_LENGTH = 36  # UUID length for string representation

BaseDBModelType = TypeVar("BaseDBModelType", bound="BaseDBModel")  # pylint: disable=invalid-name


class Base(AsyncAttrs, DeclarativeBase):
    """
    Base class for all SQLAlchemy declarative models in the project.
    """

    pass


class BaseDBModel(MappedAsDataclass, Base, kw_only=True):
    """
    Abstract base class for all database models.

    Provides common fields (uuid, created_at, updated_at) and policy-based access control.
    Subclasses can define __policies__ as a list of BasePolicy instances to control row-level access.
    """

    __abstract__ = True
    __policies__: ClassVar[list["BasePolicy"]] = []

    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default_factory=now_utc, index=True)
    updated_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default_factory=now_utc, onupdate=now_utc)

    uuid: Mapped[str] = mapped_column(
        String(UUID_LENGTH),
        primary_key=True,
        default_factory=lambda: str(uuid4()),
        index=True,
    )

    @classmethod
    def fqpn(cls) -> str:
        """
        Fully Qualified Path Name (FQPN) for the object.
        """
        return f"{cls.__module__}.{cls.__name__}"

    @classmethod
    async def apply_policies_to_query(
        cls: type[BaseDBModelType], query: Select[Any], user_uuid: str | None, policies: list[BasePolicy] | None = None
    ) -> Select[Any]:
        """
        Returns the filter statement for use in a where clause.
        Returns None if no policy filters exist.
        """

        permissive_filters = [
            await p.filter(user_uuid, cls)
            for p in (policies or cls.__policies__)
            if p.policy_type.value == "permissive"
        ]
        if not permissive_filters:
            # if no permissive filters exist we don't give any access
            # we also don't need to evaluate restrictive filters, because nobody has access anyway
            # we log a warning, because this is probably not intended
            logger.warning(
                "No permissive filters found for %s. No access granted. Returning empty result set.", cls.__name__
            )
            return query.where(cls.uuid == None)  # noqa: E711

        restrictive_filters = [
            await p.filter(user_uuid, cls)
            for p in (policies or cls.__policies__)
            if p.policy_type.value == "restrictive"
        ]

        if restrictive_filters:
            where_clause = and_(and_(*restrictive_filters), or_(*permissive_filters))  # noqa: E711
        else:
            where_clause = or_(*permissive_filters)

        return query.where(where_clause)

    @classmethod
    async def get_one_or_none_with_policies(
        cls: type[BaseDBModelType],
        _filter_clause: TypeWhereClause | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> BaseDBModelType | None:
        """
        Retrieve a single database row matching all provided key-value pairs, with custom filtering.
        This method applies all restrictive (AND) and permissive (OR) policies defined in __policies__ of the model.

        Filtering options:
        - Additional key-value filters can be passed as kwargs (combined with AND).
        - If _filter_clause is provided, it is added as an extra SQLAlchemy filter.

        Returns None if no match is found.
        """
        return await cls.get_one_or_none(
            _apply_policies=True,
            _filter_clause=_filter_clause,
            **kwargs,
        )

    @classmethod
    async def get_one_or_none(
        cls: type[BaseDBModelType],
        _apply_policies: bool = False,
        _filter_clause: TypeWhereClause | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> BaseDBModelType | None:
        """
        Retrieve a single database row matching all provided key-value pairs, with optional policy and custom filtering.

        Filtering options:
        - _apply_policies: applies all restrictive (AND) and permissive (OR) policies defined in __policies__
        - Additional key-value filters can be passed as kwargs (combined with AND).
        - If _filter_clause is provided, it is added as an extra SQLAlchemy filter.

        Returns None if no match is found.
        """
        query = select(cls)
        if _apply_policies:
            query = await cls.apply_policies_to_query(query, user_uuid=auth_context.user_uuid)
        if kwargs:
            query = query.where(and_(*(getattr(cls, key) == value for key, value in kwargs.items())))
        if _filter_clause is not None:
            query = query.where(_filter_clause)
        result = await db.session.execute(query)
        return result.unique().scalars().one_or_none()

    @classmethod
    async def get_list_with_policies(
        cls: type[BaseDBModelType],
        _filter_clause: TypeWhereClause | None = None,
        _order_by: Literal[None]
        | _ColumnExpressionOrStrLabelArgument[Any]
        | list[_ColumnExpressionOrStrLabelArgument[Any]] = None,
        _per_page: int = -1,
        _page: int = 1,
        **kwargs: Any,  # noqa: ANN401
    ) -> Sequence[BaseDBModelType]:
        """
        Retrieve a list of database rows with flexible filtering, ordering, and pagination.
        This method applies all restrictive (AND) and permissive (OR) policies defined in __policies__ of the model.

        Filtering options:
        - _filter_clause: custom SQLAlchemy filter clause.
        - Additional key-value filters can be passed as kwargs (combined with AND).

        Ordering and pagination:
        - _order_by: SQLAlchemy column or list of columns to order by.
        - _per_page: If > 0, limits the number of results per page.
        - _page: Page number for paginated results (1-based).

        Returns a sequence of matching rows.
        """
        return await cls.get_list(
            _apply_policies=True,
            _filter_clause=_filter_clause,
            _order_by=_order_by,
            _per_page=_per_page,
            _page=_page,
            **kwargs,
        )

    @classmethod
    async def get_list(
        cls: type[BaseDBModelType],
        _apply_policies: bool = False,
        _filter_clause: TypeWhereClause | None = None,
        _order_by: Literal[None]
        | _ColumnExpressionOrStrLabelArgument[Any]
        | list[_ColumnExpressionOrStrLabelArgument[Any]] = None,
        _per_page: int = -1,
        _page: int = 1,
        **kwargs: Any,  # noqa: ANN401
    ) -> Sequence[BaseDBModelType]:
        """
        Retrieve a list of database rows with flexible filtering, ordering, and pagination.

        Filtering options:
        - _apply_policies: applies all restrictive (AND) and permissive (OR) policies defined in __policies__
        - Additional key-value filters can be passed as kwargs (combined with AND).
        - If _filter_clause is provided, it is added as an extra SQLAlchemy filter.

        Ordering and pagination:
        - _order_by: SQLAlchemy column or list of columns to order by.
        - _per_page: If > 0, limits the number of results per page.
        - _page: Page number for paginated results (1-based).

        Returns a sequence of matching rows.
        """
        query = select(cls)

        if _apply_policies:
            # if user_uuid is given we apply the policies
            query = await cls.apply_policies_to_query(query, user_uuid=auth_context.user_uuid)

        # filter by keys
        if kwargs:
            query = query.where(and_(*(getattr(cls, key) == value for key, value in kwargs.items())))
        # filter by custom filter clause
        if _filter_clause is not None:
            query = query.where(_filter_clause)

        if _order_by is not None:
            query = query.order_by(*_order_by if isinstance(_order_by, list) else [_order_by])
        if _per_page > 0:
            query = query.limit(_per_page).offset((_page - 1) * _per_page)

        result = await db.session.execute(query)
        return result.unique().scalars().all()

    async def on_persist_pre(self) -> None:
        """
        Hook method called before persisting the instance to the database.
        Can be overridden in subclasses for custom behavior.
        """
        pass

    async def on_persist_post(self) -> None:
        """
        Hook method called after persisting the instance to the database.
        Can be overridden in subclasses for custom behavior.
        """
        pass

    @final
    async def persist(self) -> Self:
        """
        Persist the current instance to the database and flush the session.

        Order of operations:
        1. on_persist_pre
        2. ModelPersistPreSignal
        3. persist
        4. on_persist_post
        5. ModelPersistPostSignal
        """
        await self.on_persist_pre()
        await signals.emit(ModelPersistPreSignal(obj=self))
        db.session.add(self)
        await db.session.flush()
        await self.on_persist_post()
        await signals.emit(ModelPersistPostSignal(obj=self))

        return self

    async def on_delete_pre(self) -> None:
        """
        Hook method called before deleting the instance from the database.
        Can be overridden in subclasses for custom behavior.
        """
        pass

    async def on_delete_post(self) -> None:
        """
        Hook method called after deleting the instance from the database.
        Can be overridden in subclasses for custom behavior.
        """
        pass

    @final
    async def delete(self) -> None:
        """
        Delete the current instance from the database and flush the session.

        Order of operation is:
        1. on_delete_pre
        2. ModelDeletePreSignal
        3. delete
        4. on_delete_post
        5. ModelDeletePostSignal
        """
        await self.on_delete_pre()
        await signals.emit(ModelDeletePreSignal(obj=self))
        await db.session.delete(self)
        await db.session.flush()
        await self.on_delete_post()
        await signals.emit(ModelDeletePostSignal(obj=self))

    async def give_role_permission(self, role_name: str, permission: str) -> None:
        from brickworks.core.models.role_model import role_acl_table

        dialect_name = db.session.bind.dialect.name
        if dialect_name == "postgresql":
            from sqlalchemy.dialects.postgresql import insert as pg_insert

            stmt_pg = (
                pg_insert(role_acl_table)
                .values(role_name=role_name, object_uuid=self.uuid, object_fqpn=self.fqpn(), permission=permission)
                .on_conflict_do_nothing(index_elements=["role_name", "object_uuid", "permission"])
            )
            await db.session.execute(stmt_pg)
        elif dialect_name == "sqlite":
            stmt_sqlite = (
                insert(role_acl_table)
                .values(role_name=role_name, object_uuid=self.uuid, object_fqpn=self.fqpn(), permission=permission)
                .prefix_with("OR IGNORE")
            )
            await db.session.execute(stmt_sqlite)
        else:
            NotImplementedError(f"Dialect {dialect_name} is not supported for role ACL insertion.")

    async def remove_role_permission(self, role_name: str, permission: str) -> None:
        from brickworks.core.models.role_model import role_acl_table

        stmt = (
            delete(role_acl_table)
            .where(
                and_(
                    role_acl_table.c.role_name == role_name,
                    role_acl_table.c.object_uuid == self.uuid,
                    role_acl_table.c.permission == permission,
                )
            )
            .execution_options(synchronize_session="fetch")
        )
        await db.session.execute(stmt)
        await db.session.flush()

    async def role_permissions(self, role_name: str) -> list[str]:
        from brickworks.core.models.role_model import role_acl_table

        stmt = select(role_acl_table.c.permission).where(role_acl_table.c.role_name == role_name)
        result = await db.session.execute(stmt)
        return [row[0] for row in result.all()]

    async def has_role_permission(self, role_name: str, permission: str) -> bool:
        from brickworks.core.models.role_model import role_acl_table

        stmt = select(role_acl_table.c.permission).where(
            and_(role_acl_table.c.role_name == role_name, role_acl_table.c.object_uuid == self.uuid)
        )
        result = await db.session.execute(stmt)
        return any(row[0] == permission for row in result.all())


class ModelDeletePreSignal(BaseSignal):
    """
    Signal emitted before a database row is deleted.
    """

    obj: BaseDBModel


class ModelDeletePostSignal(ModelDeletePreSignal):
    """
    Signal emitted after a database row is deleted.
    """

    pass


class ModelPersistPreSignal(BaseSignal):
    """
    Signal emitted before a database row is persisted.
    """

    obj: BaseDBModel


class ModelPersistPostSignal(ModelPersistPreSignal):
    """
    Signal emitted after a database row is persisted.
    """

    pass



from sqlalchemy import and_, select

from brickworks.core.acl.base_policy import BasePolicy, PolicyTypeEnum
from brickworks.core.models import BaseDBModel, UserModel
from brickworks.core.models.base_view import BaseView
from brickworks.core.models.role_model import role_acl_table
from brickworks.core.models.user_model import UserStatusEnum
from brickworks.core.utils.sqlalchemy import AlwaysFalseWhereClause, AlwaysTrueWhereClause, TypeWhereClause


class AllowPublicAccessPolicy(BasePolicy):
    """
    Policy that allows public access to all records of the class.
    PERMISSIVE
    """

    policy_type = PolicyTypeEnum.PERMISSIVE

    async def filter(self, user_uuid: str | None, obj_class: type["BaseDBModel"] | type["BaseView"]) -> TypeWhereClause:
        # evaluates to True for all records
        return AlwaysTrueWhereClause


class AllowActiveUserAccessPolicy(BasePolicy):
    """
    Policy that allows active users access to all records of the class.
    PERMISSIVE
    """

    policy_type = PolicyTypeEnum.PERMISSIVE

    async def filter(self, user_uuid: str | None, obj_class: type["BaseDBModel"] | type["BaseView"]) -> TypeWhereClause:
        user = await UserModel.get_one_or_none(uuid=user_uuid)
        if user and user.status == UserStatusEnum.ACTIVE:
            # if the user is active we allow access to all records
            return AlwaysTrueWhereClause

        # if user does not exist or is not active, we don't give access
        return AlwaysFalseWhereClause


class RoleAllowPolicy(BasePolicy):
    """
    Policy to give roles access to specific objects with specific permissions.
    """

    policy_type = PolicyTypeEnum.PERMISSIVE

    def __init__(self, role_name: str, permission: str) -> None:
        self.role_name = role_name
        self.permission = permission

    async def filter(self, user_uuid: str | None, obj_class: type["BaseDBModel"]) -> TypeWhereClause:
        if user_uuid is None:
            # if no user is provided, we assume public access (which has no roles)
            return AlwaysFalseWhereClause

        # first we need to check if the user has the role
        user = await UserModel.get_one_or_none(uuid=user_uuid, _apply_policies=False)
        if not user or not await user.has_role(self.role_name):
            return AlwaysFalseWhereClause

        return (
            select(1)
            .where(
                and_(
                    role_acl_table.c.role_name == self.role_name,
                    role_acl_table.c.object_uuid == obj_class.uuid,
                    role_acl_table.c.permission == self.permission,
                )
            )
            .exists()
        )


class RoleBasedAccessPolicy(BasePolicy):
    """
    Policy that gives a specific role access to all records of the class.
    By default the policy is permissive, meaning that it does not restrict access to records granted by other policies.
    If restrictive is set to True, the policy requires prevents access to all records unless the user has the
    given role.
    """

    policy_type = PolicyTypeEnum.PERMISSIVE

    def __init__(self, role_name: str, restrictive: bool = False) -> None:
        self.role_name = role_name
        if restrictive:
            self.policy_type = PolicyTypeEnum.RESTRICTIVE

    async def filter(self, user_uuid: str | None, obj_class: type["BaseDBModel"] | type["BaseView"]) -> TypeWhereClause:
        if user_uuid is None:
            # if no user is provided, we assume public access (which has no roles)
            return AlwaysFalseWhereClause
        # first we need to check if the user has the role
        user = await UserModel.get_one_or_none(uuid=user_uuid, _apply_policies=False)
        if not user or not await user.has_role(self.role_name):
            return AlwaysFalseWhereClause

        return AlwaysTrueWhereClause



import asyncio
import logging

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from brickworks.core.cache import cache
from brickworks.core.db import db
from brickworks.core.exceptions import DuplicateException
from brickworks.core.models.base_dbmodel import BaseDBModel
from brickworks.core.settings import settings

logger = logging.getLogger(__name__)


@cache.lru_cache(master_tenant=True, expire=60)
async def get_domain_schema_mapping() -> dict[str, str]:
    return await _get_domain_schema_mapping()


async def create_schema(schema_name: str) -> None:
    """
    Runs the mason CLI to migrate the database for the given schema name asynchronously.
    """

    process = await asyncio.create_subprocess_exec(
        "mason",
        "db",
        "migrate",
        "--schema",
        schema_name,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await process.communicate()
    if process.returncode != 0:
        raise RuntimeError(f"mason db migrate failed: {stderr.decode().strip()}")
    logger.info(f"Created schema {schema_name} with mason db migrate")


async def drop_schema(schema_name: str) -> None:
    """
    Runs the mason CLI to drop the database schema for the given schema name asynchronously.
    """

    process = await asyncio.create_subprocess_exec(
        "mason",
        "db",
        "drop",
        schema_name,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await process.communicate()
    if process.returncode != 0:
        raise RuntimeError(f"mason db drop failed: {stderr.decode().strip()}")
    logger.info(f"Dropped schema {schema_name} with mason db drop")


async def _get_domain_schema_mapping() -> dict[str, str]:
    """
    Returns a dict of domain->schema mappings.
    """
    async with db(schema=settings.MASTER_DB_SCHEMA):
        tenant_models = await TenantModel.get_list()
        domain_map = {tenant_model.domain: tenant_model.schema for tenant_model in tenant_models}
        domain_map[settings.MASTER_DOMAIN] = settings.MASTER_DB_SCHEMA
    return domain_map


class TenantModel(BaseDBModel):
    __tablename__ = "core_tenants"
    schema: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    domain: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    label: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)  # public name

    @classmethod
    async def create(cls, schema: str, domain: str, label: str) -> "TenantModel":
        if schema == settings.MASTER_DB_SCHEMA or domain == settings.MASTER_DOMAIN:
            raise DuplicateException("Tenant already exists")
        new_tenant = await TenantModel(schema=schema, domain=domain, label=label).persist()

        # clear the cache after creating a new tenant, but only after the session is committed
        db.add_post_commit_callback(get_domain_schema_mapping.cache_clear)
        await create_schema(schema)
        return new_tenant

    async def on_delete_post(self) -> None:
        # clear the cache after deleting a tenant, but only after the session is committed
        db.add_post_commit_callback(get_domain_schema_mapping.cache_clear)

        # also drop the schema associated with this tenant
        async def drop_tenant_schema() -> None:
            await drop_schema(self.schema)

        db.add_post_commit_callback(drop_tenant_schema)
